"""
This module contains the main :class:`.TRAKer` class, which is the front-facing
class for TRAK. See the `README <https://github.com/MadryLab/trak>`_ and `docs
<https://trak.readthedocs.io/en/latest/>`_ for example usage.

In short, methods of the :class:`.TRAKer` class are used to compute TRAK scores
for a set of model checkpoints, a set of target samples, and a set of train samples.
This is done in two stages:
- Featurizing. :func:`.TRAKer.featurize` and :func:`.TRAKer.finalize_features`
  are used to compute the TRAK features for a set of model checkpoints and a set
  of train samples.
- Scoring. :func:`.TRAKer.start_scoring_checkpoint`, :func:`.TRAKer.score`, and
  :func:`.TRAKer.finalize_scores` are used to compute the TRAK scores for a set
  of target samples, given the TRAK features computed in the previous step.

"""
from .modelout_functions import AbstractModelOutput, TASK_TO_MODELOUT
from .projectors import (
    ProjectionType,
    AbstractProjector,
    CudaProjector,
    BasicProjector,
    ChunkedCudaProjector,
)
from .gradient_computers import FunctionalGradientComputer, AbstractGradientComputer
from .score_computers import AbstractScoreComputer, BasicScoreComputer
from .savers import AbstractSaver, MmapSaver, ModelIDException
from .utils import get_num_params, get_parameter_chunk_sizes

from typing import Iterable, Optional, Union
from pathlib import Path
from tqdm import tqdm
from torch import Tensor

import logging
import numpy as np
import torch

ch = torch


class TRAKer:
    """The main front-facing class for TRAK. See the `README
    <https://github.com/MadryLab/trak>`_ and `docs
    <https://trak.readthedocs.io/en/latest/>`_ for example usage.

    """

    def __init__(
        self,
        model: torch.nn.Module,
        task: Union[AbstractModelOutput, str],
        train_set_size: int,
        save_dir: str = "./trak_results",
        load_from_save_dir: bool = True,
        device: Union[str, torch.device] = "cuda",
        gradient_computer: AbstractGradientComputer = FunctionalGradientComputer,
        projector: Optional[AbstractProjector] = None,
        saver: Optional[AbstractSaver] = None,
        score_computer: Optional[AbstractScoreComputer] = None,
        proj_dim: int = 2048,
        logging_level=logging.INFO,
        use_half_precision: bool = True,
        proj_max_batch_size: int = 32,
        projector_seed: int = 0,
        grad_wrt: Optional[Iterable[str]] = None,
    ) -> None:
        """

        Args:
            model (torch.nn.Module):
                model to use for TRAK
            task (Union[AbstractModelOutput, str]):
                Type of model that TRAK will be ran on. Accepts either one of
                the following strings:
                - :code:`image_classification`
                - :code:`text_classification`
                - :code:`clip`
                or an instance of some implementation of the abstract class
                :class:`.AbstractModelOutput`.
            train_set_size (int):
                Size of the train set that TRAK is featurizing
            save_dir (str, optional):
                Directory to save final TRAK scores, intermediate results, and
                metadata. Defaults to :code:'./trak_results'.
            load_from_save_dir (bool, optional):
                If True, the :class`.TRAKer` instance will attempt to load
                existing metadata from save_dir. May lead to I/O issues if
                multiple TRAKer instances ran in parallel have this flag set to
                True. See the SLURM tutorial for more details.
            device (Union[str, torch.device], optional):
                torch device on which to do computations. Defaults to 'cuda'.
            gradient_computer (AbstractGradientComputer, optional):
                Class to use to get per-example gradients. See
                :class:`.AbstractGradientComputer` for more details. Defaults to
                :class:`.FunctionalGradientComputer`.
            projector (Optional[AbstractProjector], optional):
                Either set :code:`proj_dim` and a :class:`.CudaProjector`
                Rademacher projector will be used or give a custom subclass of
                :class:`.AbstractProjector` class and leave :code:`proj_dim` as
                None. Defaults to None.
            saver (Optional[AbstractSaver], optional):
                Class to use for saving intermediate results and final TRAK
                scores to RAM/disk. If None, the :class:`.MmapSaver` will
                be used. Defaults to None.
            score_computer (Optional[AbstractScoreComputer], optional):
                Class to use for computing the final TRAK scores. If None, the
                :class:`.BasicScoreComputer` will be used. Defaults to None.
            proj_dim (int, optional):
                Dimension of the projected TRAK features. See Section 4.3 of
                `our paper <https://arxiv.org/abs/2303.14186>`_ for more
                details. Defaults to 2048.
            logging_level (int, optional):
                Logging level for TRAK loggers. Defaults to logging.INFO.
            use_half_precision (bool, optional):
                If True, TRAK will use half precision (float16) for all
                computations and arrays will be stored in float16. Otherwise, it
                will use float32. Defaults to True.
            proj_max_batch_size (int):
                Batch size used by fast_jl if the CudaProjector is used. Must be
                a multiple of 8. The maximum batch size is 32 for A100 GPUs, 16
                for V100 GPUs, 40 for H100 GPUs. Defaults to 32.
            projector_seed (int):
                Random seed used by the projector. Defaults to 0.
            grad_wrt (Optional[Iterable[str]], optional):
                If not None, the gradients will be computed only with respect to
                the parameters specified in this list. The list should contain
                the names of the parameters to compute gradients with respect to,
                as they appear in the model's state dictionary. If None,
                gradients are taken with respect to all model parameters.
                Defaults to None.

        """

        self.model = model
        self.task = task
        self.train_set_size = train_set_size
        self.device = device
        self.dtype = ch.float16 if use_half_precision else ch.float32
        self.grad_wrt = grad_wrt

        logging.basicConfig()
        self.logger = logging.getLogger("TRAK")
        self.logger.setLevel(logging_level)

        self.num_params = get_num_params(self.model)
        if self.grad_wrt is not None:
            d = dict(self.model.named_parameters())
            self.num_params_for_grad = sum(
                [d[param_name].numel() for param_name in self.grad_wrt]
            )
        else:
            self.num_params_for_grad = self.num_params
        # inits self.projector
        self.proj_seed = projector_seed
        self.init_projector(
            projector=projector,
            proj_dim=proj_dim,
            proj_max_batch_size=proj_max_batch_size,
        )

        # normalize to make X^TX numerically stable
        # doing this instead of normalizing the projector matrix
        self.normalize_factor = ch.sqrt(
            ch.tensor(self.num_params_for_grad, dtype=ch.float32)
        )

        self.save_dir = Path(save_dir).resolve()
        self.load_from_save_dir = load_from_save_dir

        if type(self.task) is str:
            self.task = TASK_TO_MODELOUT[self.task]()

        self.gradient_computer = gradient_computer(
            model=self.model,
            task=self.task,
            grad_dim=self.num_params_for_grad,
            dtype=self.dtype,
            device=self.device,
            grad_wrt=self.grad_wrt,
        )

        if score_computer is None:
            score_computer = BasicScoreComputer
        self.score_computer = score_computer(
            dtype=self.dtype, device=self.device, logging_level=logging_level
        )

        metadata = {
            "JL dimension": self.proj_dim,
            "JL matrix type": self.projector.proj_type,
            "train set size": self.train_set_size,
        }

        if saver is None:
            saver = MmapSaver
        self.saver = saver(
            save_dir=self.save_dir,
            metadata=metadata,
            train_set_size=self.train_set_size,
            proj_dim=self.proj_dim,
            load_from_save_dir=self.load_from_save_dir,
            logging_level=logging_level,
            use_half_precision=use_half_precision,
        )

        self.ckpt_loaded = "no ckpt loaded"

    def init_projector(
        self,
        projector: Optional[AbstractProjector],
        proj_dim: int,
        proj_max_batch_size: int,
    ) -> None:
        """Initialize the projector for a traker class

        Args:
            projector (Optional[AbstractProjector]):
                JL projector to use. If None, a CudaProjector will be used (if
                possible).
            proj_dim (int):
                Dimension of the projected gradients and TRAK features.
            proj_max_batch_size (int):
                Batch size used by fast_jl if the CudaProjector is used. Must be
                a multiple of 8. The maximum batch size is 32 for A100 GPUs, 16
                for V100 GPUs, 40 for H100 GPUs.
        """

        self.projector = projector
        if projector is not None:
            self.proj_dim = self.projector.proj_dim
            if self.proj_dim == 0:  # using NoOpProjector
                self.proj_dim = self.num_params_for_grad

        else:
            using_cuda_projector = False
            self.proj_dim = proj_dim
            if self.device == "cpu":
                self.logger.info("Using BasicProjector since device is CPU")
                projector = BasicProjector
                # Sampling from bernoulli distribution is not supported for
                # dtype float16 on CPU; playing it safe here by defaulting to
                # normal projection, rather than rademacher
                proj_type = ProjectionType.normal
                self.logger.info("Using Normal projection")
            else:
                try:
                    import fast_jl

                    test_gradient = ch.ones(1, self.num_params_for_grad).cuda()
                    num_sms = ch.cuda.get_device_properties(
                        "cuda"
                    ).multi_processor_count
                    fast_jl.project_rademacher_8(
                        test_gradient, self.proj_dim, 0, num_sms
                    )
                    projector = CudaProjector
                    using_cuda_projector = True

                except (ImportError, RuntimeError, AttributeError) as e:
                    self.logger.error(f"Could not use CudaProjector.\nReason: {str(e)}")
                    self.logger.error("Defaulting to BasicProjector.")
                    projector = BasicProjector
                proj_type = ProjectionType.rademacher

            if using_cuda_projector:
                max_chunk_size, param_chunk_sizes = get_parameter_chunk_sizes(
                    self.model, proj_max_batch_size
                )
                self.logger.debug(
                    (
                        f"the max chunk size is {max_chunk_size}, ",
                        "while the model has the following chunk sizes",
                        f"{param_chunk_sizes}.",
                    )
                )

                if (
                    len(param_chunk_sizes) > 1
                ):  # we have to use the ChunkedCudaProjector
                    self.logger.info(
                        (
                            f"Using ChunkedCudaProjector with"
                            f"{len(param_chunk_sizes)} chunks of sizes"
                            f"{param_chunk_sizes}."
                        )
                    )
                    rng = np.random.default_rng(self.proj_seed)
                    seeds = rng.integers(
                        low=0,
                        high=500,
                        size=len(param_chunk_sizes),
                    )
                    projector_per_chunk = [
                        projector(
                            grad_dim=chunk_size,
                            proj_dim=self.proj_dim,
                            seed=seeds[i],
                            proj_type=ProjectionType.rademacher,
                            max_batch_size=proj_max_batch_size,
                            dtype=self.dtype,
                            device=self.device,
                        )
                        for i, chunk_size in enumerate(param_chunk_sizes)
                    ]
                    self.projector = ChunkedCudaProjector(
                        projector_per_chunk,
                        max_chunk_size,
                        param_chunk_sizes,
                        proj_max_batch_size,
                        self.device,
                        self.dtype,
                    )
                    return  # do not initialize projector below

            self.logger.debug(
                f"Initializing projector with grad_dim {self.num_params_for_grad}"
            )
            self.projector = projector(
                grad_dim=self.num_params_for_grad,
                proj_dim=self.proj_dim,
                seed=self.proj_seed,
                proj_type=proj_type,
                max_batch_size=proj_max_batch_size,
                dtype=self.dtype,
                device=self.device,
            )
            self.logger.debug(f"Initialized projector with proj_dim {self.proj_dim}")

    def load_checkpoint(
        self,
        checkpoint: Iterable[Tensor],
        model_id: int,
        _allow_featurizing_already_registered=False,
    ) -> None:
        """Loads state dictionary for the given checkpoint; initializes arrays
        to store TRAK features for that checkpoint, tied to the model ID.

        Args:
            checkpoint (Iterable[Tensor]):
                state_dict to load
            model_id (int):
                a unique ID for a checkpoint
            _allow_featurizing_already_registered (bool, optional):
                Only use if you want to override the default behaviour that
                :code:`featurize` is forbidden on already registered model IDs.
                Defaults to None.

        """
        if self.saver.model_ids.get(model_id) is None:
            self.saver.register_model_id(
                model_id, _allow_featurizing_already_registered
            )
        else:
            self.saver.load_current_store(model_id)

        self.model.load_state_dict(checkpoint)
        self.model.eval()

        self.gradient_computer.load_model_params(self.model)

        self._last_ind = 0
        self.ckpt_loaded = model_id

    def featurize(
        self,
        batch: Iterable[Tensor],
        inds: Optional[Iterable[int]] = None,
        num_samples: Optional[int] = None,
    ) -> None:
        """Creates TRAK features for the given batch by computing the gradient
        of the model output function and projecting it. In the notation of the
        paper, for an input pair :math:`z=(x,y)`, model parameters
        :math:`\\theta`, and JL projection matrix :math:`P`, this method
        computes :math:`P^\\top \\nabla_\\theta f(z_i, \\theta)`.
        Additionally, this method computes the gradient of the out-to-loss
        function (in the notation of the paper, the :math:`Q` term in Section
        3.4).

        Either :code:`inds` or :code:`num_samples` must be specified. Using
        :code:`num_samples` will write sequentially into the internal store of
        the :func:`TRAKer`.

        Args:
            batch (Iterable[Tensor]):
                input batch
            inds (Optional[Iterable[int]], optional):
                Indices of the batch samples in the train set. Defaults to None.
            num_samples (Optional[int], optional):
                Number of samples in the batch. Defaults to None.

        """
        assert (
            self.ckpt_loaded == self.saver.current_model_id
        ), "Load a checkpoint using traker.load_checkpoint before featurizing"
        assert (inds is None) or (
            num_samples is None
        ), "Exactly one of num_samples and inds should be specified"
        assert (inds is not None) or (
            num_samples is not None
        ), "Exactly one of num_samples and inds should be specified"

        if num_samples is not None:
            inds = np.arange(self._last_ind, self._last_ind + num_samples)
            self._last_ind += num_samples
        else:
            num_samples = inds.reshape(-1).shape[0]

        # handle re-starting featurizing from a partially featurized model (some inds already featurized)
        _already_done = (self.saver.current_store["is_featurized"][inds] == 1).reshape(
            -1
        )
        inds = inds[~_already_done]
        if len(inds) == 0:
            self.logger.debug("All samples in batch already featurized.")
            return 0

        grads = self.gradient_computer.compute_per_sample_grad(batch=batch)
        grads = self.projector.project(grads, model_id=self.saver.current_model_id)
        grads /= self.normalize_factor
        self.saver.current_store["grads"][inds] = (
            grads.to(self.dtype).cpu().clone().detach()
        )

        loss_grads = self.gradient_computer.compute_loss_grad(batch)
        self.saver.current_store["out_to_loss"][inds] = (
            loss_grads.to(self.dtype).cpu().clone().detach()
        )

        self.saver.current_store["is_featurized"][inds] = 1
        self.saver.serialize_current_model_id_metadata()

    def finalize_features(
        self, model_ids: Iterable[int] = None, del_grads: bool = False
    ) -> None:
        """For a set of checkpoints :math:`C` (specified by model IDs), and
        gradients :math:`\\{ \\Phi_c \\}_{c\\in C}`, this method computes
        :math:`\\Phi_c (\\Phi_c^\\top\\Phi_c)^{-1}` for all :math:`c\\in C`
        and stores the results in the internal store of the :func:`TRAKer`
        class.

        Args:
            model_ids (Iterable[int], optional): A list of model IDs for which
                features should be finalized. If None, features are finalized
                for all model IDs in the :code:`save_dir` of the :class:`.TRAKer`
                class. Defaults to None.

        """

        # this method is memory-intensive, so we're freeing memory beforehand
        torch.cuda.empty_cache()
        self.projector.free_memory()

        if model_ids is None:
            model_ids = list(self.saver.model_ids.keys())

        self._last_ind = 0

        for model_id in tqdm(model_ids, desc="Finalizing features for all model IDs.."):
            if self.saver.model_ids.get(model_id) is None:
                raise ModelIDException(
                    f"Model ID {model_id} not registered, not ready for finalizing."
                )
            elif self.saver.model_ids[model_id]["is_featurized"] == 0:
                raise ModelIDException(
                    f"Model ID {model_id} not fully featurized, not ready for finalizing."
                )
            elif self.saver.model_ids[model_id]["is_finalized"] == 1:
                self.logger.warning(
                    f"Model ID {model_id} already finalized, skipping .finalize_features for it."
                )
                continue

            self.saver.load_current_store(model_id)

            g = ch.as_tensor(self.saver.current_store["grads"])
            xtx = self.score_computer.get_xtx(g)
            self.logger.debug(f"XTX is {xtx}")

            features = self.score_computer.get_x_xtx_inv(g, xtx)
            self.logger.debug(f"Features are {features}")
            self.saver.current_store["features"][:] = features.to(self.dtype).cpu()
            if del_grads:
                self.saver.del_grads(model_id)

            self.saver.model_ids[self.saver.current_model_id]["is_finalized"] = 1
            self.saver.serialize_current_model_id_metadata()

    def start_scoring_checkpoint(
        self,
        exp_name: str,
        checkpoint: Iterable[Tensor],
        model_id: int,
        num_targets: int,
    ) -> None:
        """This method prepares the internal store of the :class:`.TRAKer` class
        to start computing scores for a set of targets.

        Args:
            exp_name (str):
                Experiment name. Each experiment should have a unique name, and
                it corresponds to a set of targets being scored. The experiment
                name is used as the name for saving the target features, as well
                as scores produced by this method in the :code:`save_dir` of the
                :class:`.TRAKer` class.
            checkpoint (Iterable[Tensor]):
                model checkpoint (state dict)
            model_id (int):
                a unique ID for a checkpoint
            num_targets (int):
                number of targets to score

        """
        self.saver.init_experiment(exp_name, num_targets, model_id)

        self.model.load_state_dict(checkpoint)
        self.model.eval()
        self.gradient_computer.load_model_params(self.model)

        # TODO: make this exp_name-dependent
        # e.g. make it a value in self.saver.experiments[exp_name]
        self._last_ind_target = 0

    def score(
        self,
        batch: Iterable[Tensor],
        inds: Optional[Iterable[int]] = None,
        num_samples: Optional[int] = None,
    ) -> None:
        """This method computes the (intermediate per-checkpoint) TRAK scores
        for a batch of targets and stores them in the internal store of the
        :class:`.TRAKer` class.

        Either :code:`inds` or :code:`num_samples` must be specified. Using
        :code:`num_samples` will write sequentially into the internal store of
        the :class:`.TRAKer`.

        Args:
            batch (Iterable[Tensor]):
                input batch
            inds (Optional[Iterable[int]], optional):
                Indices of the batch samples in the train set. Defaults to None.
            num_samples (Optional[int], optional):
                Number of samples in the batch. Defaults to None.

        """
        assert (inds is None) or (
            num_samples is None
        ), "Exactly one of num_samples and inds should be specified"
        assert (inds is not None) or (
            num_samples is not None
        ), "Exactly one of num_samples and inds should be specified"

        if self.saver.model_ids[self.saver.current_model_id]["is_finalized"] == 0:
            self.logger.error(
                f"Model ID {self.saver.current_model_id} not finalized, cannot score"
            )
            return None

        if num_samples is not None:
            inds = np.arange(self._last_ind_target, self._last_ind_target + num_samples)
            self._last_ind_target += num_samples
        else:
            num_samples = inds.reshape(-1).shape[0]

        grads = self.gradient_computer.compute_per_sample_grad(batch=batch)

        grads = self.projector.project(grads, model_id=self.saver.current_model_id)
        grads /= self.normalize_factor

        exp_name = self.saver.current_experiment_name
        self.saver.current_store[f"{exp_name}_grads"][inds] = (
            grads.to(self.dtype).cpu().clone().detach()
        )

    def finalize_scores(
        self,
        exp_name: str,
        model_ids: Iterable[int] = None,
        allow_skip: bool = False,
    ) -> Tensor:
        """This method computes the final TRAK scores for the given targets,
        train samples, and model checkpoints (specified by model IDs).

        Args:
            exp_name (str):
                Experiment name. Each experiment should have a unique name, and
                it corresponds to a set of targets being scored. The experiment
                name is used as the name for saving the target features, as well
                as scores produced by this method in the :code:`save_dir` of the
                :class:`.TRAKer` class.
            model_ids (Iterable[int], optional):
                A list of model IDs for which
                scores should be finalized. If None, scores are computed
                for all model IDs in the :code:`save_dir` of the :class:`.TRAKer`
                class. Defaults to None.
            allow_skip (bool, optional):
                If True, raises only a warning, instead of an error, when target
                gradients are not computed for a given model ID. Defaults to
                False.

        Returns:
            Tensor: TRAK scores

        """
        # reset counter for inds used for scoring
        self._last_ind_target = 0

        if model_ids is None:
            model_ids = self.saver.model_ids
        else:
            model_ids = {
                model_id: self.saver.model_ids[model_id] for model_id in model_ids
            }
        assert len(model_ids) > 0, "No model IDs to finalize scores for"

        if self.saver.experiments.get(exp_name) is None:
            raise ValueError(
                f"Experiment {exp_name} does not exist. Create it\n\
                              and compute scores first before finalizing."
            )

        num_targets = self.saver.experiments[exp_name]["num_targets"]
        _completed = [False] * len(model_ids)

        self.saver.load_current_store(list(model_ids.keys())[0], exp_name, num_targets)
        _scores_mmap = self.saver.current_store[f"{exp_name}_scores"]
        _scores_on_cpu = ch.zeros(*_scores_mmap.shape, device="cpu")
        if self.device != "cpu":
            _scores_on_cpu.pin_memory()

        _avg_out_to_losses = np.zeros(
            (self.saver.train_set_size, 1),
            dtype=np.float16 if self.dtype == ch.float16 else np.float32,
        )

        for j, model_id in enumerate(
            tqdm(model_ids, desc="Finalizing scores for all model IDs..")
        ):
            self.saver.load_current_store(model_id)
            try:
                self.saver.load_current_store(model_id, exp_name, num_targets)
            except OSError as e:
                if allow_skip:
                    self.logger.warning(
                        f"Could not read target gradients for model ID {model_id}. Skipping."
                    )
                    continue
                else:
                    raise e

            if self.saver.model_ids[self.saver.current_model_id]["is_finalized"] == 0:
                self.logger.warning(
                    f"Model ID {self.saver.current_model_id} not finalized, cannot score"
                )
                continue

            g = ch.as_tensor(self.saver.current_store["features"], device=self.device)
            g_target = ch.as_tensor(
                self.saver.current_store[f"{exp_name}_grads"], device=self.device
            )

            self.score_computer.get_scores(g, g_target, accumulator=_scores_on_cpu)
            # .cpu().detach().numpy()

            _avg_out_to_losses += self.saver.current_store["out_to_loss"]
            _completed[j] = True

        _num_models_used = float(sum(_completed))

        # only write to mmap (on disk) once at the end
        _scores_mmap[:] = (_scores_on_cpu.numpy() / _num_models_used) * (
            _avg_out_to_losses / _num_models_used
        )

        self.logger.debug(f"Scores dtype is {_scores_mmap.dtype}")
        self.saver.save_scores(exp_name)
        self.scores = _scores_mmap

        return self.scores



from tqdm import tqdm
from pathlib import Path
from itertools import product
import logging
import pytest
import torch as ch

from trak import TRAKer
from trak.projectors import BasicProjector

from .utils import construct_rn9, get_dataloader, eval_correlations
from .utils import download_cifar_checkpoints, download_cifar_betons


def get_projector(use_cuda_projector, dtype):
    if use_cuda_projector:
        return None
    return BasicProjector(
        grad_dim=2273856,
        proj_dim=1024,
        seed=0,
        proj_type="rademacher",
        dtype=dtype,
        device="cuda:0",
    )


PARAM = list(
    product(
        [False, True],  # serialize
        [False, True],  # basic / cuda projector
        [ch.float16, ch.float32],  # projection dtype
        [100, 32],  # batch size
    )
)


@pytest.mark.parametrize("serialize, use_cuda_projector, dtype, batch_size", PARAM)
@pytest.mark.cuda
def test_cifar_acc(serialize, use_cuda_projector, dtype, batch_size, tmp_path):
    device = "cuda:0"
    exp_name = "test_experimet"

    projector = get_projector(use_cuda_projector, dtype)
    model = construct_rn9().to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    BETONS_PATH = Path(tmp_path).joinpath("cifar_betons")
    BETONS = download_cifar_betons(BETONS_PATH)

    loader_train = get_dataloader(BETONS, batch_size=batch_size, split="train")
    loader_val = get_dataloader(BETONS, batch_size=batch_size, split="val")

    CKPT_PATH = Path(tmp_path).joinpath("cifar_ckpts")
    ckpt_files = download_cifar_checkpoints(CKPT_PATH, "cifar2")
    ckpts = [ch.load(ckpt, map_location="cpu") for ckpt in ckpt_files]

    use_half_precision = dtype == ch.float16

    traker = TRAKer(
        model=model,
        task="image_classification",
        projector=projector,
        proj_dim=1024,
        train_set_size=10_000,
        save_dir=tmp_path,
        device=device,
        logging_level=logging.DEBUG,
        use_half_precision=use_half_precision,
    )

    for model_id, ckpt in enumerate(ckpts):
        traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
        for batch in tqdm(loader_train, desc="Computing TRAK embeddings..."):
            traker.featurize(batch=batch, num_samples=len(batch[0]))

    traker.finalize_features()

    if serialize:
        del traker
        traker = TRAKer(
            model=model,
            task="image_classification",
            projector=projector,
            proj_dim=1024,
            train_set_size=10_000,
            save_dir=tmp_path,
            device=device,
            logging_level=logging.DEBUG,
            use_half_precision=use_half_precision,
        )

    for model_id, ckpt in enumerate(ckpts):
        traker.start_scoring_checkpoint(exp_name, ckpt, model_id, num_targets=2_000)
        for batch in tqdm(loader_val, desc="Scoring..."):
            traker.score(batch=batch, num_samples=len(batch[0]))

    scores = traker.finalize_scores(exp_name)

    avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path, ds="cifar2")
    assert avg_corr > 0.062, "correlation with 3 CIFAR-2 models should be >= 0.062"



import pytest
import logging
from tqdm import tqdm
from pathlib import Path
import numpy as np
import torch as ch

from trak import TRAKer
from .utils import construct_rn9, get_dataloader, eval_correlations
from .utils import download_cifar_checkpoints, download_cifar_betons


@pytest.mark.cuda
def test_featurize_and_score_in_parallel(tmp_path):
    device = "cuda:0"
    batch_size = 100

    model = construct_rn9().to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    BETONS_PATH = Path(tmp_path).joinpath("cifar_betons")
    BETONS = download_cifar_betons(BETONS_PATH)

    loader_train = get_dataloader(BETONS, batch_size=batch_size, split="train")
    loader_val = get_dataloader(BETONS, batch_size=batch_size, split="val")

    CKPT_PATH = Path(tmp_path).joinpath("cifar_ckpts")
    ckpt_files = download_cifar_checkpoints(CKPT_PATH, ds="cifar2")
    ckpts = [ch.load(ckpt, map_location="cpu") for ckpt in ckpt_files]

    # this should be essentially equivalent to running each
    # TRAKer in a separate script
    for model_id, ckpt in enumerate(ckpts):
        traker = TRAKer(
            model=model,
            task="image_classification",
            train_set_size=10_000,
            save_dir=tmp_path,
            device=device,
            logging_level=logging.DEBUG,
        )
        traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
        for batch in tqdm(loader_train, desc="Computing TRAK embeddings..."):
            traker.featurize(batch=batch, num_samples=len(batch[0]))
        traker.finalize_features()

    for model_id, ckpt in enumerate(ckpts):
        traker = TRAKer(
            model=model,
            task="image_classification",
            train_set_size=10_000,
            save_dir=tmp_path,
            device=device,
            logging_level=logging.DEBUG,
        )

        traker.start_scoring_checkpoint(
            "test_experiment", ckpt, model_id, num_targets=2_000
        )
        for batch in tqdm(loader_val, desc="Scoring..."):
            traker.score(batch=batch, num_samples=len(batch[0]))

    scores = traker.finalize_scores(exp_name="test_experiment")

    avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path, ds="cifar2")
    assert avg_corr > 0.062, "correlation with 3 CIFAR-2 models should be >= 0.062"


@pytest.mark.cuda
def test_score_multiple(tmp_path):
    device = "cuda:0"
    batch_size = 100

    model = construct_rn9().to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    BETONS_PATH = Path(tmp_path).joinpath("cifar_betons")
    BETONS = download_cifar_betons(BETONS_PATH)

    loader_train = get_dataloader(BETONS, batch_size=batch_size, split="train")
    loader_val = get_dataloader(BETONS, batch_size=batch_size, split="val")

    CKPT_PATH = Path(tmp_path).joinpath("cifar_ckpts")
    ckpt_files = download_cifar_checkpoints(CKPT_PATH, ds="cifar2")
    ckpts = [ch.load(ckpt, map_location="cpu") for ckpt in ckpt_files]

    traker = TRAKer(
        model=model,
        task="image_classification",
        train_set_size=10_000,
        save_dir=tmp_path,
        device=device,
        logging_level=logging.DEBUG,
    )

    for model_id, ckpt in enumerate(ckpts):
        traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
        for batch in tqdm(loader_train, desc="Computing TRAK embeddings..."):
            traker.featurize(batch=batch, num_samples=len(batch[0]))
    traker.finalize_features()

    scoring_runs = range(3)
    for _ in scoring_runs:
        for model_id, ckpt in enumerate(ckpts):
            traker = TRAKer(
                model=model,
                task="image_classification",
                train_set_size=10_000,
                save_dir=tmp_path,
                device=device,
                logging_level=logging.DEBUG,
            )

            traker.start_scoring_checkpoint(
                "test_experiment", ckpt, model_id, num_targets=2_000
            )
            for batch in tqdm(loader_val, desc="Scoring..."):
                traker.score(batch=batch, num_samples=len(batch[0]))

        scores = traker.finalize_scores("test_experiment")

        avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path, ds="cifar2")
        assert avg_corr > 0.062, "correlation with 3 CIFAR-2 models should be >= 0.062"


@pytest.mark.cuda
def test_score_in_shards(tmp_path):
    device = "cuda:0"
    batch_size = 100

    model = construct_rn9().to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    BETONS_PATH = Path(tmp_path).joinpath("cifar_betons")
    BETONS = download_cifar_betons(BETONS_PATH)

    loader_train = get_dataloader(BETONS, batch_size=batch_size, split="train")

    CKPT_PATH = Path(tmp_path).joinpath("cifar_ckpts")
    ckpt_files = download_cifar_checkpoints(CKPT_PATH, ds="cifar2")
    ckpts = [ch.load(ckpt, map_location="cpu") for ckpt in ckpt_files]

    traker = TRAKer(
        model=model,
        task="image_classification",
        train_set_size=10_000,
        save_dir=tmp_path,
        device=device,
        logging_level=logging.DEBUG,
    )

    for model_id, ckpt in enumerate(ckpts):
        traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
        for batch in tqdm(loader_train, desc="Computing TRAK embeddings..."):
            traker.featurize(batch=batch, num_samples=len(batch[0]))
    traker.finalize_features()

    scoring_shards = [np.arange(1000), np.arange(1000, 2000)]
    # this should be essentially equivalent to scoring each
    # shard in a separate script
    for scoring_inds in scoring_shards:
        loader_val = get_dataloader(
            BETONS, batch_size=batch_size, split="val", indices=scoring_inds
        )
        for model_id, ckpt in enumerate(ckpts):
            traker = TRAKer(
                model=model,
                task="image_classification",
                train_set_size=10_000,
                save_dir=tmp_path,
                device=device,
                logging_level=logging.DEBUG,
            )

            traker.start_scoring_checkpoint(
                "test_experiment", ckpt, model_id, num_targets=2000
            )
            for batch_idx, batch in enumerate(tqdm(loader_val, desc="Scoring...")):
                batch_inds = scoring_inds[
                    batch_idx * batch_size : (batch_idx + 1) * batch_size
                ]
                traker.score(batch=batch, inds=batch_inds)

    scores = traker.finalize_scores("test_experiment")

    avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path, ds="cifar2")
    assert avg_corr > 0.062, "correlation with 3 CIFAR-2 models should be >= 0.062"


@pytest.mark.cuda
def test_featurize_in_shards(tmp_path):
    device = "cuda:0"
    batch_size = 100

    model = construct_rn9().to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    BETONS_PATH = Path(tmp_path).joinpath("cifar_betons")
    BETONS = download_cifar_betons(BETONS_PATH)

    loader_val = get_dataloader(BETONS, batch_size=batch_size, split="val")

    CKPT_PATH = Path(tmp_path).joinpath("cifar_ckpts")
    ckpt_files = download_cifar_checkpoints(CKPT_PATH, ds="cifar2")
    ckpts = [ch.load(ckpt, map_location="cpu") for ckpt in ckpt_files]

    # this should be essentially equivalent to featurizing each
    # shard in a separate script
    featurizing_shards = [np.arange(5000), np.arange(5000, 10_000)]
    for featurizing_inds in featurizing_shards:
        loader_train = get_dataloader(
            BETONS, batch_size=batch_size, split="train", indices=featurizing_inds
        )
        traker = TRAKer(
            model=model,
            task="image_classification",
            train_set_size=10_000,
            save_dir=tmp_path,
            device=device,
            logging_level=logging.DEBUG,
        )

        for model_id, ckpt in enumerate(ckpts):
            traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
            for batch_idx, batch in enumerate(
                tqdm(loader_train, desc="Computing TRAK embeddings")
            ):
                batch_inds = featurizing_inds[
                    batch_idx * batch_size : (batch_idx + 1) * batch_size
                ]
                traker.featurize(batch=batch, inds=batch_inds)

    traker.finalize_features()

    traker = TRAKer(
        model=model,
        task="image_classification",
        train_set_size=10_000,
        save_dir=tmp_path,
        device=device,
        logging_level=logging.DEBUG,
    )

    for model_id, ckpt in enumerate(ckpts):
        traker.start_scoring_checkpoint(
            "test_experiment", ckpt, model_id, num_targets=2_000
        )
        for batch in tqdm(loader_val, desc="Scoring..."):
            traker.score(batch=batch, num_samples=len(batch[0]))

    scores = traker.finalize_scores("test_experiment")

    avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path, ds="cifar2")
    assert avg_corr > 0.062, "correlation with 3 CIFAR-2 models should be >= 0.062"


@pytest.mark.cuda
def test_preemption(tmp_path):
    device = "cuda:0"
    batch_size = 100

    model = construct_rn9().to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    BETONS_PATH = Path(tmp_path).joinpath("cifar_betons")
    BETONS = download_cifar_betons(BETONS_PATH)

    loader_val = get_dataloader(BETONS, batch_size=batch_size, split="val")

    CKPT_PATH = Path(tmp_path).joinpath("cifar_ckpts")
    ckpt_files = download_cifar_checkpoints(CKPT_PATH, ds="cifar2")
    ckpts = [ch.load(ckpt, map_location="cpu") for ckpt in ckpt_files]

    # this should be essentially equivalent to featurizing each
    # shard in a separate script
    featurizing_shards = [np.arange(5000), np.arange(10_000)]
    for featurizing_inds in featurizing_shards:
        loader_train = get_dataloader(
            BETONS, batch_size=batch_size, split="train", indices=featurizing_inds
        )
        traker = TRAKer(
            model=model,
            task="image_classification",
            train_set_size=10_000,
            save_dir=tmp_path,
            device=device,
            logging_level=logging.DEBUG,
        )

        for model_id, ckpt in enumerate(ckpts):
            traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
            for batch_idx, batch in enumerate(
                tqdm(loader_train, desc="Computing TRAK embeddings")
            ):
                batch_inds = featurizing_inds[
                    batch_idx * batch_size : (batch_idx + 1) * batch_size
                ]
                traker.featurize(batch=batch, inds=batch_inds)

    traker.finalize_features()

    traker = TRAKer(
        model=model,
        task="image_classification",
        train_set_size=10_000,
        save_dir=tmp_path,
        device=device,
        logging_level=logging.DEBUG,
    )

    for model_id, ckpt in enumerate(ckpts):
        traker.start_scoring_checkpoint(
            "test_experiment", ckpt, model_id, num_targets=2_000
        )
        for batch in tqdm(loader_val, desc="Scoring..."):
            traker.score(batch=batch, num_samples=len(batch[0]))

    scores = traker.finalize_scores("test_experiment")

    avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path, ds="cifar2")
    assert avg_corr > 0.062, "correlation with 3 CIFAR-2 models should be >= 0.062"



from tqdm import tqdm
from torchvision import datasets
from trak import TRAKer
import logging
import pytest
import open_clip
import torch
torch.backends.cuda.enable_mem_efficient_sdp(False)


@pytest.mark.cuda
def test_mscoco(tmp_path, device="cuda:0"):
    model, _, preprocess = open_clip.create_model_and_transforms("RN50")
    model = model.to(device)
    model.eval()

    tokenizer = open_clip.get_tokenizer("RN50")

    ds_train = datasets.CocoCaptions(
        root="/path/to/coco2014/images/train2014",
        annFile="/path/to/coco2014/annotations/annotations/captions_train2014.json",
    )

    traker = TRAKer(
        model=model,
        task="clip",
        save_dir=tmp_path,
        train_set_size=len(ds_train),
        device=device,
        proj_dim=512,
        logging_level=logging.DEBUG,
    )

    traker.task.get_embeddings(
        model,
        ds_train,
        batch_size=1,
        size=600,
        embedding_dim=1024,
        preprocess_fn_img=lambda x: preprocess(x).to(device).unsqueeze(0),
        preprocess_fn_txt=lambda x: tokenizer(x[0]).to(device),
    )

    traker.load_checkpoint(model.state_dict(), model_id=0)
    for bind, (img, captions) in enumerate(tqdm(ds_train)):
        x = preprocess(img).to(device).unsqueeze(0)
        # selecting (wlog) the first out of 5 captions
        y = tokenizer(captions[0]).to(device)

        traker.featurize(batch=(x, y), num_samples=x.shape[0])
        if bind == 2:
            break



from tqdm import tqdm
from pathlib import Path
from itertools import product
import logging
import pytest
import torch
import torchvision

from trak import TRAKer
from trak.projectors import BasicProjector

from .utils import construct_rn9, download_cifar_checkpoints, eval_correlations


ch = torch


def get_dataloader(
    batch_size=256, num_workers=8, split="train", shuffle=False, augment=True
):
    if augment:
        transforms = torchvision.transforms.Compose(
            [
                torchvision.transforms.RandomHorizontalFlip(),
                torchvision.transforms.RandomAffine(0),
                torchvision.transforms.ToTensor(),
                torchvision.transforms.Normalize(
                    (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.201)
                ),
            ]
        )
    else:
        transforms = torchvision.transforms.Compose(
            [
                torchvision.transforms.ToTensor(),
                torchvision.transforms.Normalize(
                    (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.201)
                ),
            ]
        )

    is_train = split == "train"
    dataset = torchvision.datasets.CIFAR10(
        root="/tmp/cifar/", download=True, train=is_train, transform=transforms
    )

    loader = torch.utils.data.DataLoader(
        dataset=dataset, shuffle=shuffle, batch_size=batch_size, num_workers=num_workers
    )

    return loader


def get_projector(use_cuda_projector, dtype):
    if use_cuda_projector:
        return None
    return BasicProjector(
        grad_dim=2274880,
        proj_dim=2048,
        seed=0,
        proj_type="normal",
        block_size=400,
        dtype=dtype,
        device="cuda:0",
    )


# reduce the number of tests for CIFAR-10
PARAM = list(
    product(
        [False],  # serialize
        [True],  # basic / cuda projector
        [ch.float16],  # projection dtype
        [128],  # batch size
    )
)


@pytest.mark.parametrize("serialize, use_cuda_projector, dtype, batch_size", PARAM)
@pytest.mark.cuda
def test_cifar_acc(serialize, use_cuda_projector, dtype, batch_size, tmp_path):
    device = "cuda:0"
    projector = get_projector(use_cuda_projector, dtype)
    model = construct_rn9(10).to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    loader_train = get_dataloader(batch_size=batch_size, split="train", augment=False)
    loader_val = get_dataloader(batch_size=batch_size, split="val", augment=False)

    CKPT_PATH = Path(tmp_path).joinpath("cifar_ckpts")
    ckpt_files = download_cifar_checkpoints(CKPT_PATH)

    ckpts = [ch.load(ckpt, map_location="cpu") for ckpt in ckpt_files]

    traker = TRAKer(
        model=model,
        task="image_classification",
        projector=projector,
        train_set_size=50_000,
        save_dir=tmp_path,
        logging_level=logging.DEBUG,
        device=device,
    )

    for model_id, ckpt in enumerate(ckpts):
        traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
        for batch in tqdm(loader_train, desc="Computing TRAK embeddings..."):
            batch = [x.cuda() for x in batch]
            traker.featurize(batch=batch, num_samples=len(batch[0]))

    traker.finalize_features()

    if serialize:
        del traker
        traker = TRAKer(
            model=model,
            task="image_classification",
            projector=projector,
            train_set_size=50_000,
            save_dir=tmp_path,
            logging_level=logging.DEBUG,
            device=device,
        )

    for model_id, ckpt in enumerate(ckpts):
        traker.start_scoring_checkpoint(
            exp_name="test_experiment",
            checkpoint=ckpt,
            model_id=model_id,
            num_targets=10_000,
        )
        for batch in tqdm(loader_val, desc="Scoring..."):
            batch = [x.cuda() for x in batch]
            traker.score(batch=batch, num_samples=len(batch[0]))

    print(traker.saver.experiments)

    scores = traker.finalize_scores(exp_name="test_experiment")
    print(scores)
    print(scores.shape)

    avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path)
    assert (
        avg_corr > 0.05
    ), "correlation with the above 3 CIFAR-10 checkpoints should be >= 0.05"



from trak import TRAKer
from torchvision.models import resnet18
import logging
import pytest
import torch as ch
from trak.projectors import BasicProjector, NoOpProjector
from trak.modelout_functions import ImageClassificationModelOutput


@pytest.fixture
def cpu_proj():
    projector = BasicProjector(
        grad_dim=11689512, proj_dim=20, seed=0, proj_type="rademacher", device="cpu"
    )
    return projector


def test_class_init_cpu(tmp_path, cpu_proj):
    model = resnet18()
    TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        projector=cpu_proj,
        train_set_size=20,
        logging_level=logging.DEBUG,
        device="cpu",
    )


def test_class_init(tmp_path, cpu_proj):
    model = resnet18()
    TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        projector=cpu_proj,
        train_set_size=20,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )


def test_load_ckpt(tmp_path, cpu_proj):
    model = resnet18()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        projector=cpu_proj,
        train_set_size=20,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)


def test_load_ckpt_repeat(tmp_path, cpu_proj):
    model = resnet18()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        projector=cpu_proj,
        train_set_size=20,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.load_checkpoint(ckpt, model_id=1)


@pytest.mark.cuda
def test_featurize(tmp_path):
    model = resnet18().cuda().eval()
    N = 32
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)


@pytest.mark.cuda
def test_max_batch_size(tmp_path):
    model = resnet18().cuda().eval()
    N = 32
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        proj_max_batch_size=16,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)


def test_class_featurize_cpu(tmp_path, cpu_proj):
    model = resnet18()
    N = 5
    batch = ch.randn(N, 3, 32, 32), ch.randint(low=0, high=10, size=(N,))
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        projector=cpu_proj,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cpu",
    )

    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)


def test_class_featurize_noop(tmp_path):
    model = resnet18()
    N = 5
    batch = ch.randn(N, 3, 32, 32), ch.randint(low=0, high=10, size=(N,))
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        projector=NoOpProjector(device="cpu"),
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cpu",
    )

    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)


@pytest.mark.cuda
def test_forgot_loading_ckpt(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    with pytest.raises(
        AssertionError,
        match="Load a checkpoint using traker.load_checkpoint before featurizing",
    ):
        traker.featurize(batch, num_samples=N)


@pytest.mark.cuda
def test_finalize_features(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()


@pytest.mark.cuda
def test_finalize_features_multiple_ftr(tmp_path):
    model = resnet18().cuda().eval()
    N = 10
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize([x[:3] for x in batch], num_samples=3)
    traker.featurize([x[3:6] for x in batch], num_samples=3)
    traker.featurize([x[6:] for x in batch], num_samples=4)
    traker.finalize_features()


@pytest.mark.cuda
def test_finalize_features_multiple_ftr_and_id(tmp_path):
    model = resnet18().cuda().eval()
    N = 10
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    for model_id in range(2):
        traker.load_checkpoint(ckpt, model_id=model_id)
        traker.featurize([x[:3] for x in batch], num_samples=3)
        traker.featurize([x[3:6] for x in batch], num_samples=3)
        traker.featurize([x[6:] for x in batch], num_samples=4)
    traker.finalize_features()


@pytest.mark.cuda
def test_score(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()
    traker.start_scoring_checkpoint("test_experiment", ckpt, 0, num_targets=N)
    traker.score(batch, num_samples=N)


@pytest.mark.cuda
def test_score_finalize(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()

    traker.start_scoring_checkpoint("test_experiment", ckpt, 0, num_targets=N)
    traker.score(batch, num_samples=N)
    traker.finalize_scores(exp_name="test_experiment")


@pytest.mark.cuda
def test_score_finalize_some_model_ids(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)

    traker.load_checkpoint(ckpt, model_id=1)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()

    traker.start_scoring_checkpoint("test_experiment", ckpt, 0, num_targets=N)
    traker.score(batch, num_samples=N)
    traker.finalize_scores(exp_name="test_experiment", model_ids=[0])


@pytest.mark.cuda
def test_score_finalize_split(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)

    traker.load_checkpoint(ckpt, model_id=1)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()

    traker.start_scoring_checkpoint("test_experiment", ckpt, 0, num_targets=N)
    traker.score(batch, num_samples=N)

    traker.start_scoring_checkpoint("test_experiment", ckpt, 1, num_targets=N)
    traker.score(batch, num_samples=N)

    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
    )
    traker.finalize_scores(exp_name="test_experiment")


@pytest.mark.cuda
def test_score_finalize_full_precision(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
        use_half_precision=False,
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()

    traker.start_scoring_checkpoint("test_experiment", ckpt, 0, num_targets=N)
    traker.score(batch, num_samples=N)
    traker.finalize_scores(exp_name="test_experiment")


def test_custom_model_output(tmp_path, cpu_proj):
    model = resnet18()
    TRAKer(
        model=model,
        task=ImageClassificationModelOutput(),
        save_dir=tmp_path,
        projector=cpu_proj,
        train_set_size=20,
        logging_level=logging.DEBUG,
        device="cpu",
    )


def test_grad_wrt_last_layer(tmp_path):
    model = resnet18().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32), ch.randint(low=0, high=10, size=(N,))
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cpu",
        use_half_precision=False,
        grad_wrt=["fc.weight", "fc.bias"],
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()

    traker.start_scoring_checkpoint("test_experiment", ckpt, 0, num_targets=N)
    traker.score(batch, num_samples=N)
    traker.finalize_scores(exp_name="test_experiment")


@pytest.mark.cuda
def test_grad_wrt_last_layer_cuda(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(
        model=model,
        task="image_classification",
        save_dir=tmp_path,
        train_set_size=N,
        logging_level=logging.DEBUG,
        device="cuda:0",
        grad_wrt=["fc.weight", "fc.bias"],
    )
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()

    traker.start_scoring_checkpoint("test_experiment", ckpt, 0, num_targets=N)
    traker.score(batch, num_samples=N)
    traker.finalize_scores(exp_name="test_experiment")



from tqdm import tqdm
from torch.utils.data import DataLoader
import torch.nn as nn
import pytest
import logging

from trak import TRAKer

from datasets import load_dataset
from transformers import (
    AutoConfig,
    AutoModelForSequenceClassification,
    AutoTokenizer,
    default_data_collator,
)


GLUE_TASK_TO_KEYS = {
    "cola": ("sentence", None),
    "mnli": ("premise", "hypothesis"),
    "mrpc": ("sentence1", "sentence2"),
    "qnli": ("question", "sentence"),
    "qqp": ("question1", "question2"),
    "rte": ("sentence1", "sentence2"),
    "sst2": ("sentence", None),
    "stsb": ("sentence1", "sentence2"),
    "wnli": ("sentence1", "sentence2"),
}

# for testing purposes
TRAIN_SET_SIZE = 20
VAL_SET_SIZE = 10


class SequenceClassificationModel(nn.Module):
    """
    Wrapper for HuggingFace sequence classification models.
    """

    def __init__(self):
        super().__init__()
        self.config = AutoConfig.from_pretrained(
            "bert-base-cased",
            num_labels=2,
            finetuning_task="qnli",
            cache_dir=None,
            revision="main",
            use_auth_token=None,
        )

        self.model = AutoModelForSequenceClassification.from_pretrained(
            "bert-base-cased",
            config=self.config,
            cache_dir=None,
            revision="main",
            use_auth_token=None,
            ignore_mismatched_sizes=False,
        )

        self.model.eval().cuda()

    def forward(self, input_ids, token_type_ids, attention_mask):
        return self.model(
            input_ids=input_ids,
            token_type_ids=token_type_ids,
            attention_mask=attention_mask,
        ).logits


def get_dataset(split, inds=None):
    raw_datasets = load_dataset(
        "glue",
        "qnli",
        cache_dir=None,
        use_auth_token=None,
    )
    sentence1_key, sentence2_key = GLUE_TASK_TO_KEYS["qnli"]

    tokenizer = AutoTokenizer.from_pretrained(
        "bert-base-cased",
        cache_dir=None,
        use_fast=True,
        revision="main",
        use_auth_token=False,
    )

    padding = "max_length"
    max_seq_length = 128

    def preprocess_function(examples):
        # Tokenize the texts
        args = (
            (examples[sentence1_key],)
            if sentence2_key is None
            else (examples[sentence1_key], examples[sentence2_key])
        )
        result = tokenizer(
            *args, padding=padding, max_length=max_seq_length, truncation=True
        )

        return result

    raw_datasets = raw_datasets.map(
        preprocess_function,
        batched=True,
        load_from_cache_file=(not False),
        desc="Running tokenizer on dataset",
    )

    if split == "train":
        train_dataset = raw_datasets["train"]
        ds = train_dataset
    else:
        eval_dataset = raw_datasets["validation"]
        ds = eval_dataset
    return ds


def init_loaders(batch_size=10):
    ds_train = get_dataset("train")
    ds_train = ds_train.select(range(TRAIN_SET_SIZE))
    ds_val = get_dataset("val")
    ds_val = ds_val.select(range(VAL_SET_SIZE))
    return DataLoader(
        ds_train, batch_size=batch_size, shuffle=False, collate_fn=default_data_collator
    ), DataLoader(
        ds_val, batch_size=batch_size, shuffle=False, collate_fn=default_data_collator
    )


def process_batch(batch):
    return (
        batch["input_ids"],
        batch["token_type_ids"],
        batch["attention_mask"],
        batch["labels"],
    )


# model too large to test on CPU
@pytest.mark.cuda
def test_qnli(tmp_path, device="cuda"):
    loader_train, loader_val = init_loaders()

    # no need to load model from checkpoint, just testing featurization and scoring
    model = SequenceClassificationModel()

    logger = logging.getLogger("QNLI")
    logger.setLevel(logging.DEBUG)
    logger.info(f"Initializing TRAKer with device {device}")

    traker = TRAKer(
        model=model,
        task="text_classification",
        train_set_size=TRAIN_SET_SIZE,
        save_dir=tmp_path,
        device=device,
        logging_level=logging.DEBUG,
        proj_dim=512,
    )

    logger.info("Loading checkpoint")
    traker.load_checkpoint(model.state_dict(), model_id=0)
    logger.info("Loaded checkpoint")
    for batch in tqdm(loader_train, desc="Featurizing.."):
        # process batch into compatible form for TRAKer TextClassificationModelOutput
        batch = process_batch(batch)
        batch = [x.to(device) for x in batch]
        traker.featurize(batch=batch, num_samples=batch[0].shape[0])

    traker.finalize_features()

    traker.start_scoring_checkpoint(
        exp_name="qnli",
        checkpoint=model.state_dict(),
        model_id=0,
        num_targets=VAL_SET_SIZE,
    )
    for batch in tqdm(loader_val, desc="Scoring.."):
        batch = process_batch(batch)
        batch = [x.to(device) for x in batch]
        traker.score(batch=batch, num_samples=batch[0].shape[0])

    traker.finalize_scores(exp_name="qnli")



import pytest
from itertools import product

import torch as ch

try:
    import fast_jl
except ModuleNotFoundError:
    print("No fast_jl available!")

from assertpy import assert_that

bs_error_str = "CUDA error: too many resources requested for launch\nCUDA kernel errors might be asynchronously reported at some other API call, so the stacktrace below might be incorrect.\nFor debugging consider passing CUDA_LAUNCH_BLOCKING=1.\nCompile with `TORCH_USE_CUDA_DSA` to enable device-side assertions.\n"  # noqa

new_bs_error_str = f"The batch size of the CudaProjector is too large for your GPU. Reduce it by using the max_batch_size argument of the CudaProjector.\nOriginal error: {bs_error_str}"  # noqa


PARAM = list(product([8], [1024, 2048], [512, 1024, 2048], [0, 1]))


@pytest.mark.parametrize("bs, input_size, output_size, seed ", PARAM)
@pytest.mark.cuda
def test_shape(bs: int, input_size: int, output_size: int, seed: int):
    print(output_size)
    input_data = ch.ones((bs, input_size), dtype=ch.float16, device="cuda:0")

    num_sms = ch.cuda.get_device_properties(
        ch.cuda.current_device()
    ).multi_processor_count

    try:
        result = fast_jl.project_rademacher_8(input_data, output_size, seed, num_sms)
    except RuntimeError as e:
        if str(e) == bs_error_str:
            raise RuntimeError(new_bs_error_str)
        else:
            raise e

    assert_that(result.shape).is_equal_to((bs, output_size))


@pytest.mark.cuda
def test_running():
    bs = 8
    input_size = 256
    seed = 17
    output_size = 512
    input_data = ch.ones((bs, input_size), dtype=ch.float16, device="cuda:0")

    num_sms = ch.cuda.get_device_properties(
        ch.cuda.current_device()
    ).multi_processor_count

    try:
        result = fast_jl.project_rademacher_8(input_data, output_size, seed, num_sms)
    except RuntimeError as e:
        if str(e) == bs_error_str:
            raise RuntimeError(new_bs_error_str)
        else:
            raise e

    print(result.sum())


@pytest.mark.cuda
def test_even():
    bs = 8
    input_size = 10240
    seed = 64
    output_size = 1024
    input_data = ch.ones((bs, input_size), dtype=ch.float16, device="cuda:0")

    num_sms = ch.cuda.get_device_properties(
        ch.cuda.current_device()
    ).multi_processor_count

    try:
        result = fast_jl.project_rademacher_8(input_data, output_size, seed, num_sms)
    except RuntimeError as e:
        if str(e) == bs_error_str:
            raise RuntimeError(new_bs_error_str)
        else:
            raise e

    assert_that(ch.all(result % 2 == 0)).is_true()


@pytest.mark.cuda
def test_odd():
    bs = 8
    input_size = 10241
    seed = 78
    output_size = 2048
    input_data = ch.ones((bs, input_size), dtype=ch.float16, device="cuda:0")

    num_sms = ch.cuda.get_device_properties(
        ch.cuda.current_device()
    ).multi_processor_count

    try:
        result = fast_jl.project_rademacher_8(input_data, output_size, seed, num_sms)
    except RuntimeError as e:
        if str(e) == bs_error_str:
            raise RuntimeError(new_bs_error_str)
        else:
            raise e

    assert_that(ch.all(result % 2 == 1)).is_true()



import pytest
import math
from itertools import product
import numpy as np
import torch
from torch import testing

from trak.projectors import CudaProjector, ProjectionType, ChunkedCudaProjector

ch = torch


def get_max_chunk_size(
    batch_size: int,
):
    max_chunk_size = np.iinfo(np.uint32).max // batch_size
    return max_chunk_size


def make_input(
    input_shape, max_chunk_size, device="cuda", dtype=torch.float32, g_tensor=None
):
    if g_tensor is None:
        g = testing.make_tensor(*input_shape, device=device, dtype=dtype)
    else:
        g = g_tensor
    _, num_params = input_shape
    num_chunks = np.ceil(num_params / max_chunk_size).astype("int32")
    g_chunks = ch.chunk(g, num_chunks, dim=1)
    result = {}
    for i, x in enumerate(g_chunks):
        result[i] = x
        print(f"Input param group {i} shape: {x.shape}")

    return result


BasicProjector = CudaProjector

MAX_BATCH_SIZE = 32
PARAM = list(
    product(
        [0, 1, 10**8],  # seed
        [ProjectionType.normal, ProjectionType.rademacher],  # proj type
        [ch.float16, ch.float32],  # dtype
        [
            (1, 25),
            (8, 10_000),
            (16, 10_002),
            (9, 10_002),
            (16, 10_001),
            (45, 1049),
            (1, int(1e9)),
        ],  # input shape
        [4096, 1024],  # proj dim
    )
)

PARAM = list(
    product(
        [123],  # seed
        [ProjectionType.rademacher],  # proj type
        [ch.float32],  # dtype
        [
            # tests for MAXINT32 overflow
            (8, 180645096),  # pass: np.prod(shape) < np.iinfo(np.int32).max
            (31, 180645096),  # fail: np.prod(shape) > np.iinfo(np.int32).max
            (32, 180645096),  # fail: np.prod(shape) > np.iinfo(np.int32).max
            (2, 780645096),  # fail: np.prod(shape) > np.iinfo(np.int32).max
        ],  # input shape
        [15_360],  # proj dim
    )
)


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_seed_consistency(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that re-running the same projection with the same seed
    leads to the same result.
    """

    proj = BasicProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )
    batch_size = input_shape[0]
    max_chunk_size = get_max_chunk_size(batch_size)
    g = make_input(input_shape, max_chunk_size, "cuda:0", dtype)

    result = proj.project(g, model_id=0)
    result_again = proj.project(g, model_id=0)
    testing.assert_close(result, result_again, equal_nan=True)

    del g
    torch.cuda.empty_cache()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_seed_consistency_2(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that re-initializing the class and re-running the same projection
    with the same seed leads to the same result.
    """

    batch_size = input_shape[0]
    max_chunk_size = get_max_chunk_size(batch_size)
    g = make_input(input_shape, max_chunk_size, "cuda:0", dtype)

    proj = BasicProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )

    result = proj.project(g, model_id=0)

    proj_again = BasicProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )
    result_again = proj_again.project(g, model_id=0)
    testing.assert_close(result, result_again, equal_nan=True)

    del g
    torch.cuda.empty_cache()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_norm_preservation(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that norms of differences are approximately preserved.
    """
    batch_size = input_shape[0]
    max_chunk_size = get_max_chunk_size(batch_size)
    g = make_input(input_shape, max_chunk_size, "cuda:0", dtype)

    rng = np.random.default_rng(seed)
    seeds = rng.integers(
        low=0,
        high=500,
        size=len(g),
    )

    param_chunk_sizes = [v.size(1) for v in g.values()]
    projector_per_chunk = [
        BasicProjector(
            grad_dim=chunk_size,
            proj_dim=proj_dim,
            seed=seeds[i],
            proj_type=proj_type,
            max_batch_size=MAX_BATCH_SIZE,
            dtype=dtype,
            device="cuda:0",
        )
        for i, chunk_size in enumerate(param_chunk_sizes)
    ]
    proj = ChunkedCudaProjector(
        projector_per_chunk,
        max_chunk_size,
        param_chunk_sizes,
        batch_size,
        "cuda:0",
        dtype,
    )

    p = proj.project(g, model_id=0)

    delta = 0.05
    eps = np.sqrt(np.log(1 / delta) / proj_dim)
    num_trials = 100
    num_successes = 0

    # flatten
    g = ch.cat([v for v in g.values()], dim=1)
    print(f"Flattened input shape: {g.shape}")

    for _ in range(num_trials):
        i, j = np.random.choice(range(g.shape[0]), size=2)
        n = (g[i] - g[j]).norm()
        # assuming for the test that the norm of each column
        # of the projection matrix has norm sqrt(n)
        # (true for rademacher and approx true for gaussian)
        pn = (p[i] - p[j]).norm() / np.sqrt(input_shape[-1])
        res = (n - pn).cpu().abs().item()
        # 35 is an arbitrary constant
        # if NaN, just give up and count as success
        if math.isinf(res):
            print("aaaaaa")
        num_successes += int(res <= 35 * eps * n)
    assert num_successes >= num_trials * (1 - 3 * delta)  # leeway with 2 *

    del g
    torch.cuda.empty_cache()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_prod_preservation(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that dot products are approximately preserved.
    """
    batch_size = input_shape[0]
    max_chunk_size = get_max_chunk_size(batch_size)
    g = make_input(input_shape, max_chunk_size, "cuda:0", dtype)

    proj = BasicProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )

    # check that things break with a garbage matrix
    # (making sure the constant 15 is reasonable)
    # proj.proj_matrix = ch.empty_like(proj.proj_matrix)

    p = proj.project(g, model_id=0)

    delta = 0.2
    eps = np.sqrt(np.log(1 / delta) / proj_dim)
    num_trials = 100
    num_successes = 0

    # flatten
    g = ch.cat([v for v in g.values()], dim=1)
    print(f"Flattened input shape: {g.shape}")

    for _ in range(num_trials):
        i, j = np.random.choice(range(g.shape[0]), size=2)
        n = g[i] @ g[j]
        pn = (p[i] / np.sqrt(input_shape[-1])) @ (p[j] / input_shape[-1])
        res = (n.abs() - pn.abs()).cpu().abs().item()
        t = (50 * np.sqrt(proj_dim) * eps * n).abs().item()
        # if NaN, just give up and count as success
        num_successes += max(int(res <= t), math.isinf(res), math.isinf(t))

    assert num_successes >= num_trials * (1 - 2 * delta)

    del g
    torch.cuda.empty_cache()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_single_nonzero_feature(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that output takes into account every feature.
    """

    batch_size = input_shape[0]
    max_chunk_size = get_max_chunk_size(batch_size)
    g = make_input(input_shape, max_chunk_size, "cuda:0", dtype)
    for k in g.keys():
        g[k] = ch.zeros_like(g[k])

    for ind in range(input_shape[0]):
        param_group = np.random.choice(range(len(g.keys())))
        coord = np.random.choice(range(g[param_group].size(1)))
        val = ch.randn(1)
        g[param_group][ind, coord] = val.item()

    proj = BasicProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )
    p = proj.project(g, model_id=0)
    assert (~ch.isclose(p, ch.zeros_like(p))).all().item()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_first_nonzero_feature(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that output takes into account first features.
    """
    g = ch.zeros(*input_shape, device="cuda:0", dtype=dtype)
    g[:, 0] = 1.0

    batch_size = input_shape[0]
    max_chunk_size = get_max_chunk_size(batch_size)
    g = make_input(input_shape, max_chunk_size, g_tensor=g)
    print(g[0])

    proj = BasicProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )
    p = proj.project(g, model_id=0)
    assert (~ch.isclose(p, ch.zeros_like(p))).all().item()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_last_nonzero_feature(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that output takes into account last features.
    """
    g = ch.zeros(*input_shape, device="cuda:0", dtype=dtype)
    g[:, -1] = 1.0

    batch_size = input_shape[0]
    max_chunk_size = get_max_chunk_size(batch_size)
    g = make_input(input_shape, max_chunk_size, g_tensor=g)
    print(g[0])

    proj = BasicProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )
    p = proj.project(g, model_id=0)
    assert (~ch.isclose(p, ch.zeros_like(p))).all().item()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_same_features(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that output is the same for the same features
    """
    g = testing.make_tensor(*input_shape, device="cuda:0", dtype=dtype)
    g[-1] = g[0]

    batch_size = input_shape[0]
    max_chunk_size = get_max_chunk_size(batch_size)
    g = make_input(input_shape, max_chunk_size, g_tensor=g)
    for i in range(len(g)):
        print(g[i][0] == g[i][-1])

    rng = np.random.default_rng(seed)
    seeds = rng.integers(
        low=0,
        high=500,
        size=len(g),
    )

    param_chunk_sizes = [v.size(1) for v in g.values()]
    projector_per_chunk = [
        BasicProjector(
            grad_dim=chunk_size,
            proj_dim=proj_dim,
            seed=seeds[i],
            proj_type=proj_type,
            max_batch_size=MAX_BATCH_SIZE,
            dtype=dtype,
            device="cuda:0",
        )
        for i, chunk_size in enumerate(param_chunk_sizes)
    ]
    proj = ChunkedCudaProjector(
        projector_per_chunk,
        max_chunk_size,
        param_chunk_sizes,
        batch_size,
        "cuda:0",
        dtype,
    )
    p = proj.project(g, model_id=0)

    assert ch.allclose(p[0], p[-1])

    del g
    torch.cuda.empty_cache()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_orthogonality(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that orthgonality of inputs is approximately presereved whp
    """
    if input_shape[0] == 1:
        pass
    else:
        proj = BasicProjector(
            grad_dim=input_shape[-1],
            proj_dim=proj_dim,
            proj_type=proj_type,
            seed=seed,
            device="cuda:0",
            dtype=dtype,
            max_batch_size=MAX_BATCH_SIZE,
        )

        num_successes = 0
        num_trials = 10
        for _ in range(num_trials):
            g = testing.make_tensor(*input_shape, device="cuda:0", dtype=dtype)
            g[-1] -= g[0] @ g[-1] / (g[0].norm() ** 2) * g[0]

            batch_size = input_shape[0]
            max_chunk_size = get_max_chunk_size(batch_size)
            g = make_input(input_shape, max_chunk_size, g_tensor=g)

            p = proj.project(g, model_id=0)
            if p[0] @ p[-1] < 1e-3:
                num_successes += 1
        assert num_successes > 0.33 * num_trials

    del g
    torch.cuda.empty_cache()



from torch import Tensor
import tempfile
import torch
import numpy as np

ch = torch


def test_install(use_fast_jl: bool = True):
    try:
        from trak import TRAKer
    except ImportError:
        raise ImportError(
            "TRAK is not installed! Please install it using `pip install traker`"
        )

    data = (ch.randn(20, 256), ch.randint(high=2, size=(20,)))
    model = ch.nn.Linear(256, 2, bias=False)

    if use_fast_jl:
        with tempfile.TemporaryDirectory() as tmpdirname:
            data = [x.cuda() for x in data]
            model = model.cuda()
            traker = TRAKer(
                model=model,
                task="image_classification",
                proj_dim=512,
                save_dir=tmpdirname,
                train_set_size=20,
                logging_level=100,
            )
            traker.load_checkpoint(model.state_dict(), model_id=0)
            traker.featurize(data, num_samples=20)
            print("TRAK and fast_jl are installed correctly!")
    else:
        from trak.projectors import NoOpProjector

        with tempfile.TemporaryDirectory() as tmpdirname:
            traker = TRAKer(
                model=model,
                task="image_classification",
                train_set_size=20,
                proj_dim=512,
                save_dir=tmpdirname,
                projector=NoOpProjector(),
                device="cpu",
                logging_level=100,
            )
            traker.load_checkpoint(model.state_dict(), model_id=0)
            traker.featurize(data, num_samples=20)
            print("TRAK is installed correctly!")


def parameters_to_vector(parameters) -> Tensor:
    """
    Same as https://pytorch.org/docs/stable/generated/torch.nn.utils.parameters_to_vector.html
    but with :code:`reshape` instead of :code:`view` to avoid a pesky error.
    """
    vec = []
    for param in parameters:
        vec.append(param.reshape(-1))
    return ch.cat(vec)


def get_num_params(model: torch.nn.Module) -> int:
    return parameters_to_vector(model.parameters()).numel()


def is_not_buffer(ind, params_dict) -> bool:
    name = params_dict[ind]
    if (
        ("running_mean" in name)
        or ("running_var" in name)
        or ("num_batches_tracked" in name)
    ):
        return False
    return True


def vectorize(g, arr=None, device="cuda") -> Tensor:
    """
    records result into arr

    gradients are given as a dict :code:`(name_w0: grad_w0, ... name_wp:
    grad_wp)` where :code:`p` is the number of weight matrices. each
    :code:`grad_wi` has shape :code:`[batch_size, ...]` this function flattens
    :code:`g` to have shape :code:`[batch_size, num_params]`.
    """
    if arr is None:
        g_elt = g[list(g.keys())[0]]
        batch_size = g_elt.shape[0]
        num_params = 0
        for param in g.values():
            assert param.shape[0] == batch_size
            num_params += int(param.numel() / batch_size)
        arr = ch.empty(size=(batch_size, num_params), dtype=g_elt.dtype, device=device)

    pointer = 0
    for param in g.values():
        if len(param.shape) < 2:
            num_param = 1
            p = param.data.reshape(-1, 1)
        else:
            num_param = param[0].numel()
            p = param.flatten(start_dim=1).data

        arr[:, pointer : pointer + num_param] = p.to(device)
        pointer += num_param

    return arr


def get_output_memory(features: Tensor, target_grads: Tensor, target_dtype: type):
    output_shape = features.size(0) * target_grads.size(0)
    output_dtype_size = ch.empty((1,), dtype=target_dtype).element_size()

    return output_shape * output_dtype_size


def get_free_memory(device):
    reserved = ch.cuda.memory_reserved(device=device)
    allocated = ch.cuda.memory_allocated(device=device)

    free = reserved - allocated
    return free


def get_matrix_mult_standard(
    features: Tensor, target_grads: Tensor, target_dtype: type
):
    output = features @ target_grads.t()
    return output.clone().to(target_dtype)


def get_matrix_mult_blockwise(
    features: Tensor, target_grads: Tensor, target_dtype: type, bs: int
):
    s_features = features.shape[0]
    s_target_grads = target_grads.shape[0]

    bs = min(s_features, s_target_grads, bs)

    # Copy the data in a pinned memory location to allow non-blocking
    # copies to the GPU
    features = features.pin_memory()
    target_grads = target_grads.pin_memory()

    # precompute all the blocks we will have to compute
    slices = []
    for i in range(int(np.ceil(s_features / bs))):
        for j in range(int(np.ceil(s_target_grads / bs))):
            slices.append((slice(i * bs, (i + 1) * bs), slice(j * bs, (j + 1) * bs)))

    # Allocate memory for the final output.
    final_output = ch.empty(
        (s_features, s_target_grads), dtype=target_dtype, device="cpu"
    )

    # Output buffers pinned on the CPU to be able to collect data from the
    # GPU asynchronously
    # For each of our (2) cuda streams we need two output buffer, one
    # is currently written on with the next batch of result and the
    # second one is already finished and getting copied on the final output

    # If the size is not a multiple of batch size we need extra buffers
    # with the proper shapes
    outputs = [
        ch.zeros((bs, bs), dtype=target_dtype, device=features.device).pin_memory()
        for x in range(4)
    ]
    left_bottom = s_features % bs
    options = [outputs]  # List of buffers we can potentially use
    if left_bottom:
        outputs_target_gradsottom = [
            ch.zeros(
                (left_bottom, bs), dtype=target_dtype, device=features.device
            ).pin_memory()
            for x in range(4)
        ]
        options.append(outputs_target_gradsottom)
    left_right = s_target_grads % bs
    if left_right:
        outputs_right = [
            ch.zeros(
                (bs, left_right), dtype=target_dtype, device=features.device
            ).pin_memory()
            for x in range(4)
        ]
        options.append(outputs_right)
    if left_right and left_bottom:
        outputs_corner = [
            ch.zeros(
                (left_bottom, left_right), dtype=target_dtype, device=features.device
            ).pin_memory()
            for x in range(4)
        ]
        options.append(outputs_corner)

    streams = [ch.cuda.Stream() for x in range(2)]

    # The slice that was computed last and need to now copied onto the
    # final output
    previous_slice = None

    def find_buffer_for_shape(shape):
        for buff in options:
            if buff[0].shape == shape:
                return buff
        return None

    for i, (slice_i, slice_j) in enumerate(slices):
        with ch.cuda.stream(streams[i % len(streams)]):
            # Copy the relevant blocks from CPU to the GPU asynchronously
            features_i = features[slice_i, :].cuda(non_blocking=True)
            target_grads_j = target_grads[slice_j, :].cuda(non_blocking=True)

            output_slice = features_i @ target_grads_j.t()

            find_buffer_for_shape(output_slice.shape)[i % 4].copy_(
                output_slice, non_blocking=False
            )

        # Write the previous batch of data from the temporary buffer
        # onto the final one (note that this was done by the other stream
        # so we swap back to the other one
        with ch.cuda.stream(streams[(i + 1) % len(streams)]):
            if previous_slice is not None:
                output_slice = final_output[previous_slice[0], previous_slice[1]]
                output_slice.copy_(
                    find_buffer_for_shape(output_slice.shape)[(i - 1) % 4],
                    non_blocking=True,
                )

        previous_slice = (slice_i, slice_j)

    # Wait for all the calculations/copies to be done
    ch.cuda.synchronize()

    # Copy the last chunk to the final result (from the appropriate buffer)
    output_slice = final_output[previous_slice[0], previous_slice[1]]
    output_slice.copy_(
        find_buffer_for_shape(output_slice.shape)[i % 4], non_blocking=True
    )

    return final_output


def get_matrix_mult(
    features: Tensor,
    target_grads: Tensor,
    target_dtype: torch.dtype = None,
    batch_size: int = 8096,
    use_blockwise: bool = False,
) -> Tensor:
    """

    Computes features @ target_grads.T. If the output matrix is too large to fit
    in memory, it will be computed in blocks.

    Args:
        features (Tensor):
            The first matrix to multiply.
        target_grads (Tensor):
            The second matrix to multiply.
        target_dtype (torch.dtype, optional):
            The dtype of the output matrix. If None, defaults to the dtype of
            features. Defaults to None.
        batch_size (int, optional):
            The batch size to use for blockwise matrix multiplication. Defaults
            to 8096.
        use_blockwise (bool, optional):
            Whether or not to use blockwise matrix multiplication. Defaults to
            False.

    """
    if target_dtype is None:
        target_dtype = features.dtype

    if use_blockwise:
        return get_matrix_mult_blockwise(
            features.cpu(), target_grads.cpu(), target_dtype, batch_size
        )
    elif features.device.type == "cpu":
        return get_matrix_mult_standard(features, target_grads, target_dtype)

    output_memory = get_output_memory(features, target_grads, target_dtype)
    free_memory = get_free_memory(features.device)

    if output_memory < free_memory:
        return get_matrix_mult_standard(features, target_grads, target_dtype)
    else:
        return get_matrix_mult_blockwise(
            features.cpu(), target_grads.cpu(), target_dtype, batch_size
        )


def get_parameter_chunk_sizes(
    model: torch.nn.Module,
    batch_size: int,
):
    """The :class:`CudaProjector` supports projecting when the product of the
    number of parameters and the batch size is less than the the max value of
    int32. This function computes the number of parameters that can be projected
    at once for a given model and batch size.

    The method returns a tuple containing the maximum number of parameters that
    can be projected at once and a list of the actual number of parameters in
    each chunk (a sequence of paramter groups).  Used in
    :class:`ChunkedCudaProjector`.
    """
    param_shapes = []
    for p in model.parameters():
        param_shapes.append(p.numel())

    param_shapes = np.array(param_shapes)

    chunk_sum = 0
    max_chunk_size = np.iinfo(np.uint32).max // batch_size
    params_per_chunk = []

    for ps in param_shapes:
        if chunk_sum + ps >= max_chunk_size:
            params_per_chunk.append(chunk_sum)
            chunk_sum = 0

        chunk_sum += ps

    if param_shapes.sum() - np.sum(params_per_chunk) > 0:
        params_per_chunk.append(param_shapes.sum() - np.sum(params_per_chunk))

    return max_chunk_size, params_per_chunk



"""
Computing features for the TRAK algorithm involves computing (and projecting)
per-sample gradients. This module contains classes that compute these
per-sample gradients. The :code:`AbstractFeatureComputer` class defines the
interface for such gradient computers. Then, we provide two implementations:
- :class:`FunctionalFeatureComputer`: A fast implementation that uses
  :code:`torch.func` to vectorize the computation of per-sample gradients, and
  thus fully levereage parallelism.
- :class:`IterativeFeatureComputer`: A more naive implementation that only uses
  native pytorch operations (i.e. no :code:`torch.func`), and computes per-sample
  gradients in a for-loop. This is often much slower than the functional
  version, but it is useful if you cannot use :code:`torch.func`, e.g., if you
  have an old version of pytorch that does not support it, or if your application
  is not supported by :code:`torch.func`.

"""
from abc import ABC, abstractmethod
from typing import Iterable, Optional
from torch import Tensor
from .utils import get_num_params, parameters_to_vector
from .modelout_functions import AbstractModelOutput
import logging
import torch

ch = torch


class AbstractGradientComputer(ABC):
    """Implementations of the GradientComputer class should allow for
    per-sample gradients.  This is behavior is enabled with three methods:

    - the :meth:`.load_model_params` method, well, loads model parameters. It can
      be as simple as a :code:`self.model.load_state_dict(..)`

    - the :meth:`.compute_per_sample_grad` method computes per-sample gradients
      of the chosen model output function with respect to the model's parameters.

    - the :meth:`.compute_loss_grad` method computes the gradients of the loss
      function with respect to the model output (which should be a scalar) for
      every sample.

    """

    @abstractmethod
    def __init__(
        self,
        model: torch.nn.Module,
        task: AbstractModelOutput,
        grad_dim: Optional[int] = None,
        dtype: Optional[torch.dtype] = torch.float16,
        device: Optional[torch.device] = "cuda",
    ) -> None:
        """Initializes attributes, nothing too interesting happening.

        Args:
            model (torch.nn.Module):
                model
            task (AbstractModelOutput):
                task (model output function)
            grad_dim (int, optional):
                Size of the gradients (number of model parameters). Defaults to
                None.
            dtype (torch.dtype, optional):
                Torch dtype of the gradients. Defaults to torch.float16.
            device (torch.device, optional):
                Torch device where gradients will be stored. Defaults to 'cuda'.

        """
        self.model = model
        self.modelout_fn = task
        self.grad_dim = grad_dim
        self.dtype = dtype
        self.device = device

    @abstractmethod
    def load_model_params(self, model) -> None:
        ...

    @abstractmethod
    def compute_per_sample_grad(self, batch: Iterable[Tensor]) -> Tensor:
        ...

    @abstractmethod
    def compute_loss_grad(self, batch: Iterable[Tensor], batch_size: int) -> Tensor:
        ...


class FunctionalGradientComputer(AbstractGradientComputer):
    def __init__(
        self,
        model: torch.nn.Module,
        task: AbstractModelOutput,
        grad_dim: int,
        dtype: torch.dtype,
        device: torch.device,
        grad_wrt: Optional[Iterable[str]] = None,
    ) -> None:
        """Initializes attributes, and loads model parameters.

        Args:
            grad_wrt (list[str], optional):
                A list of parameter names for which to keep gradients.  If None,
                gradients are taken with respect to all model parameters.
                Defaults to None.
        """
        super().__init__(model, task, grad_dim, dtype, device)
        self.model = model
        self.num_params = get_num_params(self.model)
        self.load_model_params(model)
        self.grad_wrt = grad_wrt
        self.logger = logging.getLogger("GradientComputer")

    def load_model_params(self, model) -> None:
        """Given a a torch.nn.Module model, inits/updates the (functional)
        weights and buffers. See https://pytorch.org/docs/stable/func.html
        for more details on :code:`torch.func`'s functional models.

        Args:
            model (torch.nn.Module):
                model to load

        """
        self.func_weights = dict(model.named_parameters())
        self.func_buffers = dict(model.named_buffers())

    def compute_per_sample_grad(self, batch: Iterable[Tensor]) -> Tensor:
        """Uses functorch's :code:`vmap` (see
        https://pytorch.org/functorch/stable/generated/functorch.vmap.html#functorch.vmap
        for more details) to vectorize the computations of per-sample gradients.

        Doesn't use :code:`batch_size`; only added to follow the abstract method
        signature.

        Args:
            batch (Iterable[Tensor]):
                batch of data

        Returns:
            Tensor:
                gradients of the model output function of each sample in the
                batch with respect to the model's parameters.

        """
        # taking the gradient wrt weights (second argument of get_output, hence argnums=1)
        grads_loss = torch.func.grad(
            self.modelout_fn.get_output, has_aux=False, argnums=1
        )

        # map over batch dimensions (hence 0 for each batch dimension, and None for model params)
        grads = torch.func.vmap(
            grads_loss,
            in_dims=(None, None, None, *([0] * len(batch))),
            randomness="different",
        )(self.model, self.func_weights, self.func_buffers, *batch)

        if self.grad_wrt is not None:
            for param_name in list(grads.keys()):
                if param_name not in self.grad_wrt:
                    del grads[param_name]
        return grads

    def compute_loss_grad(self, batch: Iterable[Tensor]) -> Tensor:
        """Computes the gradient of the loss with respect to the model output

        .. math::

            \\partial \\ell / \\partial \\text{(model output)}

        Note: For all applications we considered, we analytically derived the
        out-to-loss gradient, thus avoiding the need to do any backward passes
        (let alone per-sample grads). If for your application this is not feasible,
        you'll need to subclass this and modify this method to have a structure
        similar to the one of :meth:`FunctionalGradientComputer:.get_output`,
        i.e. something like:

        .. code-block:: python

            grad_out_to_loss = grad(self.model_out_to_loss_grad, ...)
            grads = vmap(grad_out_to_loss, ...)
            ...

        Args:
            batch (Iterable[Tensor]):
                batch of data

        """
        return self.modelout_fn.get_out_to_loss_grad(
            self.model, self.func_weights, self.func_buffers, batch
        )


class IterativeGradientComputer(AbstractGradientComputer):
    def __init__(
        self,
        model,
        task: AbstractModelOutput,
        grad_dim: int,
        dtype: torch.dtype,
        device: torch.device,
        grad_wrt: Optional[Iterable[str]] = None,
    ) -> None:
        super().__init__(model, task, grad_dim, dtype, device)
        self.load_model_params(model)
        self.grad_wrt = grad_wrt
        self.logger = logging.getLogger("GradientComputer")
        if self.grad_wrt is not None:
            self.logger.warning(
                "IterativeGradientComputer: ignoring grad_wrt argument."
            )

    def load_model_params(self, model) -> Tensor:
        self.model = model
        self.model_params = list(self.model.parameters())

    def compute_per_sample_grad(self, batch: Iterable[Tensor]) -> Tensor:
        """Computes per-sample gradients of the model output function This
        method does not leverage vectorization (and is hence much slower than
        its equivalent in :class:`.FunctionalGradientComputer`). We recommend
        that you use this only if :code:`torch.func` is not available to you,
        e.g. if you have a (very) old version of pytorch.
        Args:
            batch (Iterable[Tensor]):
                batch of data
        Returns:
            Tensor:
                gradients of the model output function of each sample in the
                batch with respect to the model's parameters.
        """
        batch_size = batch[0].shape[0]
        grads = ch.zeros(batch_size, self.grad_dim).to(batch[0].device)

        margin = self.modelout_fn.get_output(self.model, *batch)
        for ind in range(batch_size):
            grads[ind] = parameters_to_vector(
                ch.autograd.grad(margin[ind], self.model_params, retain_graph=True)
            )
        return grads

    def compute_loss_grad(self, batch: Iterable[Tensor]) -> Tensor:
        """Computes the gradient of the loss with respect to the model output
        .. math::
            \\partial \\ell / \\partial \\text{(model output)}
        Note: For all applications we considered, we analytically derived the
        out-to-loss gradient, thus avoiding the need to do any backward passes
        (let alone per-sample grads). If for your application this is not feasible,
        you'll need to subclass this and modify this method to have a structure
        similar to the one of :meth:`.IterativeGradientComputer.get_output`,
        i.e. something like:
        .. code-block:: python
            out_to_loss = self.model_out_to_loss(...)
            for ind in range(batch_size):
                grads[ind] = torch.autograd.grad(out_to_loss[ind], ...)
            ...
        Args:
            batch (Iterable[Tensor]):
                batch of data
        """
        return self.modelout_fn.get_out_to_loss_grad(self.model, batch)



from tqdm import tqdm
from torch.utils.data import DataLoader
from torchvision import datasets, models, transforms
import logging
import pytest

from trak import TRAKer
from trak.projectors import BasicProjector


def test_cifar10(tmp_path, device="cpu"):
    model = models.resnet18(weights="DEFAULT")
    model.to(device)
    model.eval()

    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))]
    )

    ds_train = datasets.CIFAR10(
        root="/tmp", download=True, train=True, transform=transform
    )
    loader_train = DataLoader(ds_train, batch_size=10, shuffle=False)
    if device == "cpu":
        # the default CudaProjector does not work on cpu
        projector = BasicProjector(
            grad_dim=sum(x.numel() for x in model.parameters()),
            proj_dim=20,
            seed=0,
            proj_type="rademacher",
            device=device,
        )
    else:
        projector = None
    traker = TRAKer(
        model=model,
        task="image_classification",
        train_set_size=len(ds_train),
        projector=projector,
        save_dir=tmp_path,
        logging_level=logging.DEBUG,
        device=device,
    )

    ckpts = [model.state_dict(), model.state_dict()]
    for model_id, ckpt in enumerate(ckpts):
        traker.load_checkpoint(ckpt, model_id=model_id)

        for batch in tqdm(loader_train, desc="Computing TRAK embeddings..."):
            batch = [x.to(device) for x in batch]
            traker.featurize(batch=batch, num_samples=loader_train.batch_size)
            break  # a CPU pass takes too long lol


@pytest.mark.cuda
def test_cifar10_cuda(tmp_path):
    test_cifar10(tmp_path, device="cuda:0")



"""
This module contains classes that save TRAK results, intermediate values, and
metadata to disk. The :code:`AbstractSaver` class defines the interface for
savers. Then, we provide one implementation:
- :class:`MmapSaver`: A saver that uses memory-mapped numpy arrays. This makes
  loading and saving small chunks of data (e.g.) during featurizing feasible
  without loading the entire file into memory.
"""
from abc import ABC, abstractmethod
from typing import Optional, Iterable, Union
from pathlib import Path
import os
import logging
import json
import numpy as np
from numpy.lib.format import open_memmap
import torch

ch = torch


class AbstractSaver(ABC):
    """
    Implementations of Saver class must implement getters and setters for TRAK
    features and scores, as well as intermediate values like gradients and
    "out-to-loss-gradient".

    The Saver class also handles the recording of metadata associated with each
    TRAK run. For example, hyperparameters like "JL dimension" -- the dimension
    used for the dimensionality reduction step of TRAK (Johnson-Lindenstrauss
    projection).
    """

    @abstractmethod
    def __init__(
        self,
        save_dir: Union[Path, str],
        metadata: Iterable,
        load_from_save_dir: bool,
        logging_level: int,
        use_half_precision: bool,
    ) -> None:
        """Creates the save directory if it doesn't already exist.
        If the save directory already exists, it validates that the current
        TRAKer class has the same hyperparameters (metadata) as the one
        specified in the save directory. Next, this method loads any existing
        computed results / intermediate values in the save directory. Last, it
        initalizes the self.current_store attributes which will be later
        populated with data for the "current" model ID of the TRAKer instance.

        Args:
            save_dir (Union[Path, str]): directory to save TRAK results,
                intermediate values, and metadata
            metadata (Iterable): a dictionary containing metadata related to the
                TRAKer class
            load_from_save_dir (bool): If True, the Saver instance will attempt
                to load existing metadata from save_dir. May lead to I/O issues
                if multiple Saver instances ran in parallel have this flag set
                to True. See the SLURM tutorial in our docs for more details.
            logging_level (int):
                logging level for the logger associated with this Saver instance
            use_half_precision (bool):
                If True, the Saver instance will save all results and intermediate
                values in half precision (float16).
        """
        self.metadata = metadata
        self.save_dir = Path(save_dir).resolve()
        self.load_from_save_dir = load_from_save_dir
        self.use_half_precision = use_half_precision

        os.makedirs(self.save_dir, exist_ok=True)
        os.makedirs(self.save_dir.joinpath("scores"), exist_ok=True)

        self.logger = logging.getLogger("STORE")
        self.logger.setLevel(logging_level)

        # init TRAKer metadata
        self.metadata_file = self.save_dir.joinpath("metadata.json")
        if os.path.exists(self.metadata_file) and self.load_from_save_dir:
            with open(self.metadata_file, "r") as f:
                existsing_metadata = json.load(f)
            existing_jl_dim = int(existsing_metadata["JL dimension"])
            assert (
                self.metadata["JL dimension"] == existing_jl_dim
            ), f"In {self.save_dir} there are models using JL dimension {existing_jl_dim},\n\
                   and this TRAKer instance uses JL dimension {self.metadata['JL dimension']}."

            existing_matrix_type = existsing_metadata["JL matrix type"]
            assert (
                self.metadata["JL matrix type"] == existing_matrix_type
            ), f"In {self.save_dir} there are models using a {existing_matrix_type} JL matrix,\n\
                   and this TRAKer instance uses a {self.metadata['JL matrix type']} JL matrix."

            assert (
                self.metadata["train set size"] == existsing_metadata["train set size"]
            ), f"In {self.save_dir} there are models TRAKing\n\
                   {existsing_metadata['train set size']} examples, and in this TRAKer instance\n\
                   there are {self.metadata['train set size']} examples."

        elif self.load_from_save_dir:
            with open(self.metadata_file, "w") as f:
                json.dump(self.metadata, f)

        self.model_ids = {}
        self.experiments = {}
        self.experiments_file = self.save_dir.joinpath("experiments.json")
        if self.load_from_save_dir:
            # check if there are existing model ids in the save_dir
            self.model_ids_files = self.save_dir.rglob("id_*.json")

            for existing_model_id_file in self.model_ids_files:
                with open(existing_model_id_file, "r") as f:
                    existing_id = json.load(f)
                    existing_id = {
                        int(model_id): metadata
                        for model_id, metadata in existing_id.items()
                    }
                self.model_ids.update(existing_id)

            if os.path.isfile(self.experiments_file):
                with open(self.experiments_file, "r") as f:
                    self.experiments.update(json.load(f))
            else:
                with open(self.experiments_file, "w") as f:
                    json.dump({}, f)

        existing_ids = list(self.model_ids.keys())
        if len(existing_ids) > 0:
            self.logger.info(
                f"Existing model IDs in {self.save_dir}: {sorted(existing_ids)}"
            )
            ids_finalized = sorted(
                list([id for id, v in self.model_ids.items() if v["is_finalized"] == 1])
            )
            if len(ids_finalized) > 0:
                self.logger.info(f"Model IDs that have been finalized: {ids_finalized}")
            else:
                self.logger.info(
                    f"No model IDs in {self.save_dir} have been finalized."
                )
        else:
            self.logger.info(f"No existing model IDs in {self.save_dir}.")

        if len(list(self.experiments.keys())) > 0:
            self.logger.info("Existing TRAK scores:")
            for exp_name, values in self.experiments.items():
                self.logger.info(f"{exp_name}: {values['scores_path']}")
        else:
            self.logger.info(f"No existing TRAK scores in {self.save_dir}.")

        self.current_model_id = None
        self.current_store = {
            "grads": None,
            "out_to_loss": None,
            "features": None,
        }

    @abstractmethod
    def register_model_id(self, model_id: int) -> None:
        """Create metadata for a new model ID (checkpoint).

        Args:
            model_id (int):
                a unique ID for a checkpoint

        """
        ...

    @abstractmethod
    def serialize_current_model_id_metadata(self) -> None:
        """Write to disk / commit any updates to the metadata associated
        to the current model ID

        """
        ...

    @abstractmethod
    def init_store(self, model_id: int) -> None:
        """Initializes store for a given model ID (checkpoint).

        Args:
            model_id (int):
                a unique ID for a checkpoint
        """
        ...

    @abstractmethod
    def init_experiment(self, model_id: int) -> None:
        """Initializes store for a given experiment & model ID (checkpoint).

        Args:
            model_id (int):
                a unique ID for a checkpoint
        """
        ...

    @abstractmethod
    def load_current_store(self, model_id: int) -> None:
        """Populates the self.current_store attributes with data for the
        given model ID (checkpoint).

        Args:
            model_id (int):
                a unique ID for a checkpoint

        """
        ...

    @abstractmethod
    def save_scores(self, exp_name: str) -> None:
        """Saves scores for a given experiment name

        Args:
            exp_name (str):
                experiment name

        """
        ...

    @abstractmethod
    def del_grads(self, model_id: int, target: bool) -> None:
        """Delete the intermediate values (gradients) for a given model id

        Args:
            model_id (int):
                a unique ID for a checkpoint
            target (bool):
                if True, delete the gradients of the target samples, otherwise
                delete the train set gradients.

        """
        ...


class ModelIDException(Exception):
    """A minimal custom exception for errors related to model IDs"""

    pass


class MmapSaver(AbstractSaver):
    """A saver that uses memory-mapped numpy arrays. This makes small reads and
    writes (e.g.) during featurizing feasible without loading the entire file
    into memory.

    """

    def __init__(
        self,
        save_dir,
        metadata,
        train_set_size,
        proj_dim,
        load_from_save_dir,
        logging_level,
        use_half_precision,
    ) -> None:
        super().__init__(
            save_dir=save_dir,
            metadata=metadata,
            load_from_save_dir=load_from_save_dir,
            logging_level=logging_level,
            use_half_precision=use_half_precision,
        )
        self.train_set_size = train_set_size
        self.proj_dim = proj_dim

    def register_model_id(
        self, model_id: int, _allow_featurizing_already_registered: bool
    ) -> None:
        """This method
        1) checks if the model ID already exists in the save dir
        2) if yes, it raises an error since model IDs must be unique
        3) if not, it creates a metadata file for it and initalizes store mmaps

        Args:
            model_id (int):
                a unique ID for a checkpoint

        Raises:
            ModelIDException:
                raised if the model ID to be registered already exists

        """
        self.current_model_id = model_id

        if self.current_model_id in self.model_ids.keys() and (
            not _allow_featurizing_already_registered
        ):
            err_msg = f"model id {self.current_model_id} is already registered. Check {self.save_dir}"
            raise ModelIDException(err_msg)
        self.model_ids[self.current_model_id] = {"is_featurized": 0, "is_finalized": 0}

        self.init_store(self.current_model_id)
        self.serialize_current_model_id_metadata(already_exists=False)

    def serialize_current_model_id_metadata(self, already_exists=True) -> None:
        is_featurized = int(
            self.current_store["is_featurized"].sum() == self.train_set_size
        )

        # update the metadata JSON file
        content = {
            self.current_model_id: {
                "is_featurized": is_featurized,
                "is_finalized": self.model_ids[self.current_model_id]["is_finalized"],
            }
        }
        # update the metadata dict within the class instance
        self.model_ids[self.current_model_id]["is_featurized"] = is_featurized
        if (is_featurized == 1) or not already_exists:
            with open(
                self.save_dir.joinpath(f"id_{self.current_model_id}.json"), "w"
            ) as f:
                json.dump(content, f)

    def init_store(self, model_id) -> None:
        prefix = self.save_dir.joinpath(str(model_id))
        if os.path.exists(prefix):
            self.logger.info(f"Model ID folder {prefix} already exists")
        os.makedirs(prefix, exist_ok=True)
        featurized_so_far = np.zeros(shape=(self.train_set_size,), dtype=np.int32)
        ft = self._load(
            prefix.joinpath("_is_featurized.mmap"),
            shape=(self.train_set_size,),
            mode="w+",
            dtype=np.int32,
        )
        ft[:] = featurized_so_far[:]
        ft.flush()

        self.load_current_store(model_id, mode="w+")

    def init_experiment(self, exp_name, num_targets, model_id) -> None:
        prefix = self.save_dir.joinpath(str(model_id))
        if not os.path.exists(prefix):
            raise ModelIDException(
                f"model ID folder {prefix} does not exist,\n\
            cannot start scoring"
            )
        self.experiments[exp_name] = {
            "num_targets": num_targets,
            "scores_path": str(self.save_dir.joinpath(f"scores/{exp_name}.mmap")),
            "scores_finalized": 0,
        }

        # update experiments.json
        with open(self.experiments_file, "r") as fp:
            exp_f = json.load(fp)

        exp_f[exp_name] = self.experiments[exp_name]
        with open(self.experiments_file, "w") as fp:
            json.dump(exp_f, fp)

        if os.path.exists(prefix.joinpath(f"{exp_name}_grads.mmap")):
            mode = "r+"
        else:
            mode = "w+"
        self.load_current_store(
            model_id=model_id, exp_name=exp_name, exp_num_targets=num_targets, mode=mode
        )

    def _load(self, fname, shape, mode, dtype=None):
        if mode == "w+":
            self.logger.debug(f"Creating {fname}.")
        else:
            self.logger.debug(f"Loading {fname}.")
        if dtype is None:
            dtype = np.float16 if self.use_half_precision else np.float32
        try:
            return open_memmap(filename=fname, mode=mode, shape=shape, dtype=dtype)
        except OSError:
            self.logger.info(f"{fname} does not exist, skipping.")
            return None

    def load_current_store(
        self,
        model_id: int,
        exp_name: Optional[str] = None,
        exp_num_targets: Optional[int] = -1,
        mode: Optional[str] = "r+",
    ) -> None:
        """This method uses numpy memmaps for serializing the TRAK results and
        intermediate values.

        Args:
            model_id (int):
                a unique ID for a checkpoint
            exp_name (str, optional):
                Experiment name for which to load the features. If None, loads
                the train (source) features for a model ID. Defaults to None.
            exp_num_targets (int, optional):
                Number of targets for the experiment. Specify only when exp_name
                is not None. Defaults to -1.
            mode (str, optional):
                Defaults to 'r+'.

        """
        self.current_model_id = model_id
        if exp_name is not None:
            self.current_experiment_name = exp_name
        prefix = self.save_dir.joinpath(str(self.current_model_id))

        if exp_name is None:
            to_load = {
                "grads": (
                    prefix.joinpath("grads.mmap"),
                    (self.train_set_size, self.proj_dim),
                    None,
                ),
                "out_to_loss": (
                    prefix.joinpath("out_to_loss.mmap"),
                    (self.train_set_size, 1),
                    None,
                ),
                "features": (
                    prefix.joinpath("features.mmap"),
                    (self.train_set_size, self.proj_dim),
                    None,
                ),
                "is_featurized": (
                    prefix.joinpath("_is_featurized.mmap"),
                    (self.train_set_size, 1),
                    np.int32,
                ),
            }
        else:
            to_load = {
                f"{exp_name}_grads": (
                    prefix.joinpath(f"{exp_name}_grads.mmap"),
                    (exp_num_targets, self.proj_dim),
                    None,
                ),
                f"{exp_name}_scores": (
                    self.save_dir.joinpath(f"scores/{exp_name}.mmap"),
                    (self.train_set_size, exp_num_targets),
                    None,
                ),
            }

        for name, (path, shape, dtype) in to_load.items():
            self.current_store[name] = self._load(path, shape, mode, dtype)

    def save_scores(self, exp_name):
        assert self.current_experiment_name == exp_name
        prefix = self.save_dir.joinpath("scores")
        self.logger.info(f"Saving scores in {prefix}/{exp_name}.mmap")
        self.current_store[f"{exp_name}_scores"].flush()
        self.experiments[exp_name]["scores_finalized"] = 1
        with open(self.experiments_file, "w") as fp:
            json.dump(self.experiments, fp)

    def del_grads(self, model_id):
        grads_file = self.save_dir.joinpath(str(model_id)).joinpath("grads.mmap")

        # delete grads memmap
        grads_file.unlink()



import pytest
from itertools import product
import torch as ch
from torch import testing

from trak.projectors import CudaProjector, ProjectionType

MAX_BATCH_SIZE = 32

# TEST CASES 1
PARAM = list(
    product(
        [123],  # seed
        [ProjectionType.rademacher],  # proj type
        [ch.float32],  # dtype
        [
            (8, 180645096),  # pass: np.prod(shape) < np.iinfo(np.int32).max
            (16, 180645096),  # pass: np.prod(shape) > np.iinfo(np.int32).max
            (31, 180645096),  # fail: np.prod(shape) > np.iinfo(np.int32).max
            (32, 180645096),  # fail: np.prod(shape) > np.iinfo(np.int32).max
            (33, 180645096),  # pass: np.prod(shape) > np.iinfo(np.int32).max
            (48, 180645096),  # pass: np.prod(shape) > np.iinfo(np.int32).max
            (50, 180645096),  # pass: np.prod(shape) > np.iinfo(np.int32).max
        ],  # input shape
        [15_360],  # proj dim
    )
)

# TEST CASES 2
PARAM = list(
    product(
        [123],  # seed
        [ProjectionType.rademacher],  # proj type
        [ch.float32],  # dtype
        [
            (1, 780645096),  # pass: np.prod(shape) < np.iinfo(np.int32).max
            (5, 780645096),  # pass: np.prod(shape) > np.iinfo(np.int32).max
            (6, 780645096),  # pass: np.prod(shape) > np.iinfo(np.int32).max
            (7, 780645096),  # fail: np.prod(shape) > np.iinfo(np.int32).max
            (8, 780645096),  # fail: np.prod(shape) > np.iinfo(np.int32).max
        ],  # input shape
        [4_096],  # proj dim
    )
)


# TEST CASES 3 (ONLY for test_same_features_diff_sms)
PARAM = list(
    product(
        [123],  # seed
        [ProjectionType.rademacher],  # proj type
        [ch.float32],  # dtype
        [
            (32, 100_000),
        ],  # input shape
        [4_096],  # proj dim
    )
)


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_same_features(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that output is the same for the same features
    """
    g = testing.make_tensor(*input_shape, device="cuda:0", dtype=dtype)
    g[-1] = g[0]

    proj = CudaProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )
    p = proj.project(g, model_id=0)

    assert ch.allclose(p[0], p[-1])


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_same_features_diff_sms(
    seed,
    proj_type,
    dtype,
    proj_dim,
    input_shape,
):
    """
    Check that output is the same for the same features
    """
    g = testing.make_tensor(*input_shape, device="cuda:0", dtype=dtype)

    # project with all SMs available
    proj_full_sms = CudaProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )
    p_full_sms = proj_full_sms.project(g, model_id=0)

    # project with half SMs available
    proj_half_sms = CudaProjector(
        grad_dim=input_shape[-1],
        proj_dim=proj_dim,
        proj_type=proj_type,
        seed=seed,
        device="cuda:0",
        dtype=dtype,
        max_batch_size=MAX_BATCH_SIZE,
    )

    proj_half_sms.num_sms = max(proj_half_sms.num_sms // 2, 1)
    p_half_sms = proj_half_sms.project(g, model_id=0)

    assert ch.allclose(p_full_sms, p_half_sms)



#!/usr/bin/env python
from setuptools import setup

setup(
    name="traker",
    version="0.3.1",
    description="TRAK: Attributing Model Behavior at Scale",
    long_description="Check https://trak.csail.mit.edu/ to learn more about TRAK",
    author="MadryLab",
    author_email="trak@mit.edu",
    license_files=("LICENSE.txt",),
    packages=["trak"],
    install_requires=[
        "torch>=2.0.0",
        "numpy",
        "tqdm",
    ],
    extras_require={
        "tests": [
            "assertpy",
            "torchvision",
            "open_clip_torch",
            "wget",
            "scipy",
            "datasets",
            "transformers",
        ],
        "fast": ["fast_jl"],
    },
    include_package_data=True,
)



"""
Projectors are used to project gradients to a lower-dimensional space. This 1) allows
us to compute TRAK scores in a *much* more efficient manner, and 2) turns out to be
act as a useful regularizer (see Appendix E.1 in our paper).

Here, we provide four implementations of the projector:
- :class:`NoOpProjector` (no-op)
- :class:`BasicSingleBlockProjector` (bare-bones, inefficient implementation)
- :class:`BasicProjector` (block-wise implementation)
- :class:`CudaProjector` (a fast implementation with a custom CUDA kernel)
"""
from abc import ABC, abstractmethod
from typing import Union
from enum import Enum
import math
from torch import Tensor
import torch

from .utils import vectorize


ch = torch


class ProjectionType(str, Enum):
    normal: str = "normal"
    rademacher: str = "rademacher"


class AbstractProjector(ABC):
    """Implementations of the Projector class must implement the
    :meth:`AbstractProjector.project` method, which takes in model gradients and
    returns
    """

    @abstractmethod
    def __init__(
        self,
        grad_dim: int,
        proj_dim: int,
        seed: int,
        proj_type: Union[str, ProjectionType],
        device: Union[str, torch.device],
    ) -> None:
        """Initializes hyperparameters for the projection.

        Args:
            grad_dim (int):
                number of parameters in the model (dimension of the gradient
                vectors)
            proj_dim (int):
                dimension after the projection
            seed (int):
                random seed for the generation of the sketching (projection)
                matrix
            proj_type (Union[str, ProjectionType]):
                the random projection (JL transform) guearantees that distances
                will be approximately preserved for a variety of choices of the
                random matrix (see e.g. https://arxiv.org/abs/1411.2404). Here,
                we provide an implementation for matrices with iid Gaussian
                entries and iid Rademacher entries.
            device (Union[str, torch.device]):
                CUDA device to use

        """
        self.grad_dim = grad_dim
        self.proj_dim = proj_dim
        self.seed = seed
        self.proj_type = proj_type
        self.device = device

    @abstractmethod
    def project(self, grads: Tensor, model_id: int) -> Tensor:
        """Performs the random projection. Model ID is included
        so that we generate different projection matrices for every
        model ID.

        Args:
            grads (Tensor): a batch of gradients to be projected
            model_id (int): a unique ID for a checkpoint

        Returns:
            Tensor: the projected gradients
        """

    def free_memory(self):
        """Frees up memory used by the projector."""


class NoOpProjector(AbstractProjector):
    """
    A projector that returns the gradients as they are, i.e., implements
    :code:`projector.project(grad) = grad`.
    """

    def __init__(
        self,
        grad_dim: int = 0,
        proj_dim: int = 0,
        seed: int = 0,
        proj_type: Union[str, ProjectionType] = "na",
        device: Union[str, torch.device] = "cuda",
        *args,
        **kwargs,
    ) -> None:
        super().__init__(grad_dim, proj_dim, seed, proj_type, device)

    def project(self, grads: Tensor, model_id: int) -> Tensor:
        """A no-op method.

        Args:
            grads (Tensor): a batch of gradients to be projected
            model_id (int): a unique ID for a checkpoint

        Returns:
            Tensor: the (non-)projected gradients
        """
        return vectorize(grads, device=self.device)

    def free_memory(self):
        """A no-op method."""
        pass


class BasicSingleBlockProjector(AbstractProjector):
    """
    A bare-bones, inefficient implementation of the projection, which simply
    calls torch's matmul for the projection step.

    Note: for most model sizes (e.g. even for ResNet18), and small projection
    dimensions (e.g. anything > 100) this method will OOM on an A100.

    Unless you have a good reason to use this class (I cannot think of one, I
    added this only for testing purposes), use instead the CudaProjector or
    BasicProjector.
    """

    def __init__(
        self,
        grad_dim: int,
        proj_dim: int,
        seed: int,
        proj_type: ProjectionType,
        device,
        dtype=ch.float32,
        model_id=0,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(grad_dim, proj_dim, seed, proj_type, device)

        self.model_id = model_id
        self.proj_type = proj_type
        self.generator = ch.Generator(device=self.device)
        self.generator = self.generator.manual_seed(
            self.seed + int(1e4) * self.model_id
        )
        self.dtype = dtype

        self.proj_matrix = ch.empty(
            self.grad_dim, self.proj_dim, dtype=self.dtype, device=self.device
        )

        self.proj_matrix_available = True

        self.generate_sketch_matrix()  # updates self.proj_matrix

    def free_memory(self):
        del self.proj_matrix
        self.proj_matrix_available = False

    def generate_sketch_matrix(self):
        if not self.proj_matrix_available:
            self.proj_matrix = ch.empty(
                self.grad_dim, self.proj_dim, dtype=self.dtype, device=self.device
            )
            self.proj_matrix_available = True

        if self.proj_type == ProjectionType.normal or self.proj_type == "normal":
            self.proj_matrix.normal_(generator=self.generator)
        elif (
            self.proj_type == ProjectionType.rademacher
            or self.proj_type == "rademacher"
        ):
            self.proj_matrix.bernoulli_(p=0.5, generator=self.generator)
            # going from Bernoulli {0, 1} to Rademacher {-1, 1}
            self.proj_matrix *= 2.0
            self.proj_matrix -= 1.0
        else:
            raise KeyError(f"Projection type {self.proj_type} not recognized.")

    def project(self, grads: Tensor, model_id: int) -> Tensor:
        grads = vectorize(grads, device=self.device)
        grads = grads.to(dtype=self.dtype)
        if model_id != self.model_id:
            self.model_id = model_id
            self.generator = self.generator.manual_seed(
                self.seed + int(1e4) * self.model_id
            )
            self.generate_sketch_matrix()  # updates self.proj_matrix

        return grads @ self.proj_matrix


class BasicProjector(AbstractProjector):
    """
    A simple block-wise implementation of the projection. The projection matrix
    is generated on-device in blocks. The accumulated result across blocks is
    returned.

    Note: This class will be significantly slower and have a larger memory
    footprint than the CudaProjector. It is recommended that you use this method
    only if the CudaProjector is not available to you -- e.g. if you don't have
    a CUDA-enabled device with compute capability >=7.0 (see
    https://developer.nvidia.com/cuda-gpus).
    """

    def __init__(
        self,
        grad_dim: int,
        proj_dim: int,
        seed: int,
        proj_type: ProjectionType,
        device: torch.device,
        block_size: int = 100,
        dtype: torch.dtype = ch.float32,
        model_id=0,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(grad_dim, proj_dim, seed, proj_type, device)

        self.block_size = min(self.proj_dim, block_size)
        self.num_blocks = math.ceil(self.proj_dim / self.block_size)
        self.dtype = dtype
        self.proj_type = proj_type
        self.model_id = model_id

        self.proj_matrix = ch.empty(
            self.grad_dim, self.block_size, dtype=self.dtype, device=self.device
        )

        self.proj_matrix_available = True

        self.generator = ch.Generator(device=self.device)

        self.get_generator_states()
        self.generate_sketch_matrix(self.generator_states[0])

    def free_memory(self):
        del self.proj_matrix
        self.proj_matrix_available = False

    def get_generator_states(self):
        self.generator_states = []
        self.seeds = []
        self.jl_size = self.proj_matrix.numel()

        for i in range(self.num_blocks):
            s = self.seed + int(1e3) * i + int(1e5) * self.model_id
            self.seeds.append(s)
            self.generator = self.generator.manual_seed(s)
            self.generator_states.append(self.generator.get_state())

    def generate_sketch_matrix(self, generator_state):
        if not self.proj_matrix_available:
            self.proj_matrix = ch.empty(
                self.grad_dim, self.block_size, dtype=self.dtype, device=self.device
            )
            self.proj_matrix_available = True

        self.generator.set_state(generator_state)
        if self.proj_type == ProjectionType.normal or self.proj_type == "normal":
            self.proj_matrix.normal_(generator=self.generator)
        elif (
            self.proj_type == ProjectionType.rademacher
            or self.proj_type == "rademacher"
        ):
            self.proj_matrix.bernoulli_(p=0.5, generator=self.generator)
            self.proj_matrix *= 2.0
            self.proj_matrix -= 1.0
        else:
            raise KeyError(f"Projection type {self.proj_type} not recognized.")

    def project(self, grads: Tensor, model_id: int) -> Tensor:
        grads = vectorize(grads, device=self.device)
        grads = grads.to(dtype=self.dtype)
        sketch = ch.zeros(
            size=(grads.size(0), self.proj_dim), dtype=self.dtype, device=self.device
        )

        if model_id != self.model_id:
            self.model_id = model_id
            self.get_generator_states()  # regenerate random seeds for new model_id
            if self.num_blocks == 1:
                self.generate_sketch_matrix(self.generator_states[0])

        if self.num_blocks == 1:
            ch.matmul(grads.data, self.proj_matrix, out=sketch)
        else:
            for ind in range(self.num_blocks):
                self.generate_sketch_matrix(self.generator_states[ind])

                st = ind * self.block_size
                ed = min((ind + 1) * self.block_size, self.proj_dim)
                sketch[:, st:ed] = (
                    grads.type(self.dtype) @ self.proj_matrix[:, : (ed - st)]
                )
        return sketch.type(grads.dtype)


class CudaProjector(AbstractProjector):
    """
    A performant implementation of the projection for CUDA with compute
    capability >= 7.0.
    """

    def __init__(
        self,
        grad_dim: int,
        proj_dim: int,
        seed: int,
        proj_type: ProjectionType,
        device,
        max_batch_size: int,
        *args,
        **kwargs,
    ) -> None:
        """

        Args:
            grad_dim (int):
                Number of parameters
            proj_dim (int):
                Dimension we project *to* during the projection step
            seed (int):
                Random seed
            proj_type (ProjectionType):
                Type of randomness to use for projection matrix (rademacher or normal)
            device:
                CUDA device
            max_batch_size (int):
                Explicitly constraints the batch size the CudaProjector is going
                to use for projection. Set this if you get a 'The batch size of
                the CudaProjector is too large for your GPU' error. Must be
                either 8, 16, or 32.

        Raises:
            ValueError:
                When attempting to use this on a non-CUDA device
            ModuleNotFoundError:
                When fast_jl is not installed

        """
        super().__init__(grad_dim, proj_dim, seed, proj_type, device)
        self.max_batch_size = max_batch_size

        if isinstance(device, str):
            device = ch.device(device)

        if device.type != "cuda":
            err = "CudaProjector only works on a CUDA device; Either switch to a CUDA device, or use the BasicProjector"
            raise ValueError(err)

        self.num_sms = ch.cuda.get_device_properties(device.index).multi_processor_count

        try:
            import fast_jl

            # test run to catch at init time if projection goes through
            fast_jl.project_rademacher_8(
                ch.zeros(8, 1_000, device="cuda"), 512, 0, self.num_sms
            )
        except ImportError:
            err = "You should make sure to install the CUDA projector for traker (called fast_jl).\
                  See the installation FAQs for more details."
            raise ModuleNotFoundError(err)

    def project(
        self,
        grads: Union[dict, Tensor],
        model_id: int,
        is_grads_dict: bool = True,
    ) -> Tensor:
        if is_grads_dict:
            grads = vectorize(grads, device=self.device)
        batch_size = grads.shape[0]

        effective_batch_size = 32
        if batch_size <= 8:
            effective_batch_size = 8
        elif batch_size <= 16:
            effective_batch_size = 16

        effective_batch_size = min(self.max_batch_size, effective_batch_size)

        function_name = f"project_{self.proj_type.value}_{effective_batch_size}"
        import fast_jl

        fn = getattr(fast_jl, function_name)

        try:
            result = fn(
                grads, self.proj_dim, self.seed + int(1e4) * model_id, self.num_sms
            )
        except RuntimeError as e:
            if "CUDA error: too many resources requested for launch" in str(e):
                # provide a more helpful error message
                raise RuntimeError(
                    (
                        "The batch size of the CudaProjector is too large for your GPU. "
                        "Reduce it by using the proj_max_batch_size argument of the TRAKer.\nOriginal error:"
                    )
                )
            else:
                raise e

        return result

    def free_memory(self):
        """A no-op method."""
        pass


class ChunkedCudaProjector:
    def __init__(
        self,
        projector_per_chunk: list,
        max_chunk_size: int,
        params_per_chunk: list,
        feat_bs: int,
        device: torch.device,
        dtype: torch.dtype,
    ):
        self.projector_per_chunk = projector_per_chunk
        self.proj_dim = self.projector_per_chunk[0].proj_dim
        self.proj_type = self.projector_per_chunk[0].proj_type
        self.params_per_chunk = params_per_chunk

        self.max_chunk_size = max_chunk_size
        self.feat_bs = feat_bs
        self.device = device
        self.dtype = dtype
        self.input_allocated = False

    def allocate_input(self):
        if self.input_allocated:
            return

        self.ch_input = ch.zeros(
            size=(self.feat_bs, self.max_chunk_size),
            device=self.device,
            dtype=self.dtype,
        )

        self.input_allocated = True

    def free_memory(self):
        if not self.input_allocated:
            return

        del self.ch_input
        self.input_allocated = False

    def project(self, grads, model_id):
        self.allocate_input()
        ch_output = ch.zeros(
            size=(self.feat_bs, self.proj_dim), device=self.device, dtype=self.dtype
        )
        pointer = 0
        # iterate over params, keep a counter of params so far, and when prev
        # chunk reaches max_chunk_size, project and accumulate
        projector_index = 0
        for i, p in enumerate(grads.values()):
            if len(p.shape) < 2:
                p_flat = p.data.unsqueeze(-1)
            else:
                p_flat = p.data.flatten(start_dim=1)

            param_size = p_flat.size(1)
            if pointer + param_size > self.max_chunk_size:
                # fill remaining entries with 0
                assert pointer == self.params_per_chunk[projector_index]
                # project and accumulate
                ch_output.add_(
                    self.projector_per_chunk[projector_index].project(
                        self.ch_input[:, :pointer].contiguous(),
                        model_id=model_id,
                        is_grads_dict=False,
                    )
                )
                # reset counter
                pointer = 0
                projector_index += 1

            # continue accumulation
            actual_bs = min(self.ch_input.size(0), p_flat.size(0))
            self.ch_input[:actual_bs, pointer : pointer + param_size].copy_(p_flat)
            pointer += param_size

        # at the end, we need to project remaining items
        # fill remaining entries with 0
        assert pointer == self.params_per_chunk[projector_index]
        # project and accumulate
        ch_output[:actual_bs].add_(
            self.projector_per_chunk[projector_index].project(
                self.ch_input[:actual_bs, :pointer].contiguous(),
                model_id=model_id,
                is_grads_dict=False,
            )
        )

        return ch_output[:actual_bs]



"""
Computing scores for the TRAK algorithm from pre-computed projected gradients
involves a number of matrix multiplications. This module contains classes that
perform these operations. The :code:`AbstractScoreComputer` class defines the
interface for score computers. Then, we provide two implementations:
- :class:`BasicSingleBlockScoreComputer`: A bare-bones implementation, mostly for
    testing purposes.
- :class:`BasicScoreComputer`: A more sophisticated implementation that does
    block-wise matrix multiplications to avoid OOM errors.

"""
from abc import ABC, abstractmethod
import logging
from torch import Tensor
import torch

from .utils import get_matrix_mult

ch = torch


class AbstractScoreComputer(ABC):
    """
    The :code:`ScoreComputer` class
    Implementations of the ScoreComputer class must implement three methods:
    - :code:`get_xtx`
    - :code:`get_x_xtx_inv`
    - :code:`get_scores`
    """

    @abstractmethod
    def __init__(self, dtype, device) -> None:
        self.dtype = dtype
        self.device = device

    @abstractmethod
    def get_xtx(self, grads: Tensor) -> Tensor:
        """Computes :math:`X^\top X`, where :math:`X` is the matrix of projected
        gradients. Here, the shape of :math:`X` is :code:`(n, p)`, where
        :math:`n` is the number of training examples and :math:`p` is the
        dimension of the projection.


        Args:
            grads (Tensor): projected gradients of shape :code:`(n, p)`.

        Returns:
            Tensor: :math:`X^\top X` of shape :code:`(p, p)`.
        """

    @abstractmethod
    def get_x_xtx_inv(self, grads: Tensor, xtx: Tensor) -> Tensor:
        """Computes :math:`X(X^\top X)^{-1}`, where :math:`X` is the matrix of
        projected gradients. Here, the shape of :math:`X` is :code:`(n, p)`,
        where :math:`n` is the number of training examples and :math:`p` is the
        dimension of the projection. This function takes as input the
        pre-computed :math:`X^\top X` matrix, which is computed by the
        :code:`get_xtx` method.

        Args:
            grads (Tensor): projected gradients :math:`X` of shape :code:`(n, p)`.
            xtx (Tensor): :math:`X^\top X` of shape :code:`(p, p)`.

        Returns:
            Tensor: :math:`X(X^\top X)^{-1}` of shape :code:`(n, p)`.
        """

    @abstractmethod
    def get_scores(
        self, features: Tensor, target_grads: Tensor, accumulator: Tensor
    ) -> None:
        """Computes the scores for a given set of features and target gradients.
        In particular, this function takes in a matrix of features
        :math:`\Phi=X(X^\top X)^{-1}`, computed by the :code:`get_x_xtx_inv`
        method, and a matrix of target (projected) gradients :math:`X_{target}`.
        Then, it computes the scores as :math:`\Phi X_{target}^\top`.  The
        resulting matrix has shape :code:`(n, m)`, where :math:`n` is the number
        of training examples and :math:`m` is the number of target examples.

        The :code:`accumulator` argument is used to store the result of the
        computation. This is useful when computing scores for multiple model
        checkpoints, as it allows us to re-use the same memory for the score
        matrix.

        Args:
            features (Tensor): features :math:`\Phi` of shape :code:`(n, p)`.
            target_grads (Tensor):
                target projected gradients :math:`X_{target}` of shape
                :code:`(m, p)`.
            accumulator (Tensor): accumulator of shape :code:`(n, m)`.
        """


class BasicSingleBlockScoreComputer(AbstractScoreComputer):
    """A bare-bones implementation of :code:`ScoreComputer` that will likely
    OOM for almost all applications. Here for testing purposes only. Unless you
    have a good reason not to, you should use :func:`BasicScoreComputer`
    instead.
    """

    def get_xtx(self, grads: Tensor) -> Tensor:
        return grads.T @ grads

    def get_x_xtx_inv(self, grads: Tensor, xtx: Tensor) -> Tensor:
        # torch.linalg.inv does not support float16
        return grads @ ch.linalg.inv(xtx.float()).to(self.dtype)

    def get_scores(
        self, features: Tensor, target_grads: Tensor, accumulator: Tensor
    ) -> None:
        accumulator += (features @ target_grads.T).detach().cpu()


class BasicScoreComputer(AbstractScoreComputer):
    """An implementation of :code:`ScoreComputer` that computes matmuls in a
    block-wise manner.
    """

    def __init__(
        self,
        dtype: torch.dtype,
        device: torch.device,
        CUDA_MAX_DIM_SIZE: int = 20_000,
        logging_level=logging.INFO,
    ) -> None:
        """
        Args:
            dtype (torch.dtype):
            device (Union[str, torch.device]):
                torch device to do matmuls on
            CUDA_MAX_DIM_SIZE (int, optional):
                Size of block for block-wise matmuls. Defaults to 100_000.
            logging_level (logging level, optional):
                Logging level for the logger. Defaults to logging.info.
        """
        super().__init__(dtype, device)
        self.CUDA_MAX_DIM_SIZE = CUDA_MAX_DIM_SIZE
        self.logger = logging.getLogger("ScoreComputer")
        self.logger.setLevel(logging_level)

    def get_xtx(self, grads: Tensor) -> Tensor:
        self.proj_dim = grads.shape[1]
        result = ch.zeros(
            self.proj_dim, self.proj_dim, dtype=self.dtype, device=self.device
        )
        blocks = ch.split(grads, split_size_or_sections=self.CUDA_MAX_DIM_SIZE, dim=0)

        for block in blocks:
            result += block.T.to(self.device) @ block.to(self.device)

        return result

    def get_x_xtx_inv(self, grads: Tensor, xtx: Tensor) -> Tensor:
        blocks = ch.split(grads, split_size_or_sections=self.CUDA_MAX_DIM_SIZE, dim=0)
        xtx_inv = ch.linalg.inv(xtx.to(ch.float32))

        # center X^TX inverse a bit to avoid numerical issues when going to float16
        xtx_inv /= xtx_inv.abs().mean()

        xtx_inv = xtx_inv.to(self.dtype)

        result = ch.empty(
            grads.shape[0], xtx_inv.shape[1], dtype=self.dtype, device=self.device
        )
        for i, block in enumerate(blocks):
            start = i * self.CUDA_MAX_DIM_SIZE
            end = min(grads.shape[0], (i + 1) * self.CUDA_MAX_DIM_SIZE)
            result[start:end] = block.to(self.device) @ xtx_inv
        return result

    def get_scores(
        self, features: Tensor, target_grads: Tensor, accumulator: Tensor
    ) -> Tensor:
        train_dim = features.shape[0]
        target_dim = target_grads.shape[0]

        self.logger.debug(f"{train_dim=}, {target_dim=}")

        accumulator += (
            get_matrix_mult(features=features, target_grads=target_grads).detach().cpu()
        )
